"""
Local Field Invariants

This contains routines to compute local (`p`-adic) invariants of
quadratic forms over the rationals.
"""

# ****************************************************************************
#       Copyright (C) 2007 William Stein and Jonathan Hanke
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

###########################################################################
# TO DO: Add routines for Hasse invariants at all places, anisotropic
# places, is_semi_definite, and support for number fields.
###########################################################################

from copy import deepcopy

from sage.arith.misc import hilbert_symbol, prime_divisors
from sage.matrix.matrix_space import MatrixSpace
from sage.misc.cachefunc import cached_method
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ


def rational_diagonal_form(self, return_matrix=False):
    """
    Return a diagonal form equivalent to the given quadratic from
    over the fraction field of its defining ring.

    INPUT:

    - ``return_matrix`` -- boolean (default: ``False``); also return the
      transformation matrix

    OUTPUT: either the diagonal quadratic form `D` (if ``return_matrix`` is false)
    or the pair `(D, T)` (if ``return_matrix`` is true) where

    - ``D`` -- the diagonalized form of this quadratic form

    - ``T`` -- transformation matrix. This is such that
      ``T.transpose() * self.matrix() * T`` gives ``D.matrix()``

    Both `D` and `T` are defined over the fraction field of the
    base ring of the given form.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [0,1,-1])
        sage: Q
        Quadratic form in 2 variables over Integer Ring with coefficients:
        [ 0 1 ]
        [ * -1 ]
        sage: Q.rational_diagonal_form()
        Quadratic form in 2 variables over Rational Field with coefficients:
        [ 1/4 0 ]
        [ * -1 ]

    If we start with a diagonal form, we get back the same form defined
    over the fraction field::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q.rational_diagonal_form()
        Quadratic form in 4 variables over Rational Field with coefficients:
        [ 1 0 0 0 ]
        [ * 3 0 0 ]
        [ * * 5 0 ]
        [ * * * 7 ]

    In the following example, we check the consistency of the
    transformation matrix::

        sage: Q = QuadraticForm(ZZ, 4, range(10))
        sage: D, T = Q.rational_diagonal_form(return_matrix=True)
        sage: D
        Quadratic form in 4 variables over Rational Field with coefficients:
        [ -1/16 0 0 0 ]
        [ * 4 0 0 ]
        [ * * 13 0 ]
        [ * * * 563/52 ]
        sage: T
        [     1      0     11 149/26]
        [  -1/8      1     -2 -10/13]
        [     0      0      1 -29/26]
        [     0      0      0      1]
        sage: T.transpose() * Q.matrix() * T
        [  -1/8      0      0      0]
        [     0      8      0      0]
        [     0      0     26      0]
        [     0      0      0 563/26]
        sage: D.matrix()
        [  -1/8      0      0      0]
        [     0      8      0      0]
        [     0      0     26      0]
        [     0      0      0 563/26]

    ::

        sage: Q1 = QuadraticForm(ZZ, 4, [1, 1, 0, 0, 1, 0, 0, 1, 0, 18])
        sage: Q1
        Quadratic form in 4 variables over Integer Ring with coefficients:
        [ 1 1 0 0 ]
        [ * 1 0 0 ]
        [ * * 1 0 ]
        [ * * * 18 ]
        sage: Q1.rational_diagonal_form(return_matrix=True)
        (
        Quadratic form in 4 variables over Rational Field with coefficients:
        [ 1 0 0 0 ]
        [ * 3/4 0 0 ]
        [ * * 1 0 ]
        [ * * * 18 ]                                                         ,
        <BLANKLINE>
        [   1 -1/2    0    0]
        [   0    1    0    0]
        [   0    0    1    0]
        [   0    0    0    1]
        )

    PARI returns a singular transformation matrix for this case::

        sage: Q = QuadraticForm(QQ, 2, [1/2, 1, 1/2])
        sage: Q.rational_diagonal_form()
        Quadratic form in 2 variables over Rational Field with coefficients:
        [ 1/2 0 ]
        [ * 0 ]

    This example cannot be computed by PARI::

        sage: Q = QuadraticForm(RIF, 4, range(10))
        sage: Q.__pari__()
        Traceback (most recent call last):
        ...
        TypeError
        sage: Q.rational_diagonal_form()
        Quadratic form in 4 variables over Real Interval Field with 53 bits of precision
        with coefficients:
        [ 5 0.?e-14 0.?e-13 0.?e-13 ]
        [ * -0.05000000000000? 0.?e-12 0.?e-12 ]
        [ * * 13.00000000000? 0.?e-10 ]
        [ * * * 10.8269230769? ]

    TESTS:

    Changing the output quadratic form does not affect the caching::

        sage: Q, T = Q1.rational_diagonal_form(return_matrix=True)
        sage: Q[0,0] = 13
        sage: Q1.rational_diagonal_form()
        Quadratic form in 4 variables over Rational Field with coefficients:
        [ 1 0 0 0 ]
        [ * 3/4 0 0 ]
        [ * * 1 0 ]
        [ * * * 18 ]

    The transformation matrix is immutable::

        sage: T[0,0] = 13
        Traceback (most recent call last):
        ...
        ValueError: matrix is immutable; please change a copy instead
        (i.e., use copy(M) to change a copy of M).

    Test for a singular form::

        sage: m = matrix(GF(11), [[1,5,0,0], [5,1,9,0], [0,9,1,5], [0,0,5,1]])
        sage: qf = QuadraticForm(m)
        sage: Q, T = qf.rational_diagonal_form(return_matrix=True)
        sage: T
        [ 1  6  5 10]
        [ 0  1 10  9]
        [ 0  0  1  2]
        [ 0  0  0  1]
    """
    Q, T = self._rational_diagonal_form_and_transformation()
    T.set_immutable()

    # Quadratic forms do not support immutability, so we need to make
    # a copy to be safe.
    Q = deepcopy(Q)
    return (Q, T) if return_matrix else Q


@cached_method
def _rational_diagonal_form_and_transformation(self):
    """
    Return a diagonal form equivalent to the given quadratic from and
    the corresponding transformation matrix.

    This is over the fraction field of the base ring of the given
    quadratic form.

    OUTPUT: a tuple `(D,T)` where

    - ``D`` -- the diagonalized form of this quadratic form

    - ``T`` -- transformation matrix. This is such that
      ``T.transpose() * self.matrix() * T`` gives ``D.matrix()``

    Both `D` and `T` are defined over the fraction field of the
    base ring of the given form.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 4, [1, 1, 0, 0, 1, 0, 0, 1, 0, 18])
        sage: Q
        Quadratic form in 4 variables over Integer Ring with coefficients:
        [ 1 1 0 0 ]
        [ * 1 0 0 ]
        [ * * 1 0 ]
        [ * * * 18 ]
        sage: Q._rational_diagonal_form_and_transformation()
        (
        Quadratic form in 4 variables over Rational Field with coefficients:
        [ 1 0 0 0 ]
        [ * 3/4 0 0 ]
        [ * * 1 0 ]
        [ * * * 18 ]                                                         ,
        <BLANKLINE>
        [   1 -1/2    0    0]
        [   0    1    0    0]
        [   0    0    1    0]
        [   0    0    0    1]
        )
    """
    n = self.dim()
    K = self.base_ring().fraction_field()
    Q = self.change_ring(K)
    MS = MatrixSpace(K, n, n)

    try:
        # Try PARI if the type is supported
        pariself = self.__pari__()
        # Check that conversion back works
        MS(pariself.sage())
    except Exception:
        pass
    else:
        R = pariself.qfgaussred()
        # Diagonal matrix
        D = MS()
        for i in range(n):
            D[i, i] = R[i, i]
        newQ = Q.parent()(D)
        # Transformation matrix (inverted)
        T = MS(R.sage())
        for i in range(n):
            T[i, i] = K.one()
        try:
            return newQ, ~T
        except ZeroDivisionError:
            # Singular case is not fully supported by PARI
            pass

    # General case if conversion to/from PARI failed
    T = MS(1)

    # Clear the entries one row at a time.
    for i in range(n):

        # Deal with rows where the diagonal entry is zero.
        if Q[i, i] == 0:

            # Look for a nonzero entry and use it to make the diagonal nonzero (if it exists)
            for j in range(i + 1, n):
                if Q[i, j] != 0:
                    temp = MS(1)
                    if Q[i, j] + Q[j, j] == 0:
                        temp[j, i] = -1
                    else:
                        temp[j, i] = 1

                    # Apply the transformation
                    Q = Q(temp)
                    T = T * temp
                    break

        # Create a matrix which deals with off-diagonal entries (all at once for each row)
        temp = MS(1)
        for j in range(i + 1, n):
            if Q[i, j] != 0:
                temp[i, j] = -Q[i, j] / (Q[i, i] * 2)
                # This should only occur when Q[i,i] != 0, which the above step guarantees.

        Q = Q(temp)
        T = T * temp

    return Q, T


def signature_vector(self):
    r"""
    Return the triple `(p, n, z)` of integers where

    - `p` = number of positive eigenvalues
    - `n` = number of negative eigenvalues
    - `z` = number of zero eigenvalues

    for the symmetric matrix associated to `Q`.

    OUTPUT: a triple of integers `\geq 0`

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,0,0,-4])
        sage: Q.signature_vector()
        (1, 1, 2)

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,-3,-4])
        sage: Q.signature_vector()
        (2, 2, 0)

    ::

        sage: Q = QuadraticForm(ZZ, 4, range(10)); Q
        Quadratic form in 4 variables over Integer Ring with coefficients:
        [ 0 1 2 3 ]
        [ * 4 5 6 ]
        [ * * 7 8 ]
        [ * * * 9 ]
        sage: Q.signature_vector()
        (3, 1, 0)
    """
    diag = self.rational_diagonal_form()
    p = 0
    n = 0
    z = 0
    for i in range(diag.dim()):
        if diag[i, i] > 0:
            p += 1
        elif diag[i, i] < 0:
            n += 1
        else:
            z += 1

    return (p, n, z)


def signature(self):
    """
    Return the signature of the quadratic form, defined as:

       number of positive eigenvalues `-` number of negative eigenvalues

    of the matrix of the quadratic form.

    OUTPUT: integer

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,0,0,-4,3,11,3])
        sage: Q.signature()
        3

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,-3,-4])
        sage: Q.signature()
        0

    ::

        sage: Q = QuadraticForm(ZZ, 4, range(10)); Q
        Quadratic form in 4 variables over Integer Ring with coefficients:
        [ 0 1 2 3 ]
        [ * 4 5 6 ]
        [ * * 7 8 ]
        [ * * * 9 ]
        sage: Q.signature()
        2
    """
    p, n, _ = self.signature_vector()
    return p - n


def hasse_invariant(self, p):
    r"""
    Compute the Hasse invariant at a prime `p` or at infinity, as given on p55 of
    Cassels's book.  If `Q` is diagonal with coefficients `a_i`, then the
    (Cassels) Hasse invariant is given by

    .. MATH::

        c_p = \prod_{i < j} (a_i, a_j)_p

    where `(a,b)_p` is the Hilbert symbol at `p`.  The underlying
    quadratic form must be non-degenerate over `\QQ_p` for this to make
    sense.

    .. WARNING::

        This is different from the O'Meara Hasse invariant, which
        allows `i \leq j` in the product.  That is given by the method
        :meth:`hasse_invariant__OMeara`.

    .. NOTE::

        We should really rename this ``hasse_invariant__Cassels``, and
        set :meth:`hasse_invariant` as a front-end to it.

    INPUT:

    - ``p`` -- a prime number > 0 or `-1` for the infinite place

    OUTPUT: `1` or `-1`

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [1,2,3])
        sage: Q.rational_diagonal_form()
        Quadratic form in 2 variables over Rational Field with coefficients:
        [ 1 0 ]
        [ * 2 ]
        sage: [Q.hasse_invariant(p) for p in prime_range(20)]                           # needs sage.libs.pari
        [1, 1, 1, 1, 1, 1, 1, 1]
        sage: [Q.hasse_invariant__OMeara(p) for p in prime_range(20)]                   # needs sage.libs.pari
        [1, 1, 1, 1, 1, 1, 1, 1]

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-1])
        sage: [Q.hasse_invariant(p) for p in prime_range(20)]                           # needs sage.libs.pari
        [1, 1, 1, 1, 1, 1, 1, 1]
        sage: [Q.hasse_invariant__OMeara(p) for p in prime_range(20)]                   # needs sage.libs.pari
        [-1, 1, 1, 1, 1, 1, 1, 1]

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-1,5])
        sage: [Q.hasse_invariant(p) for p in prime_range(20)]                           # needs sage.libs.pari
        [1, 1, 1, 1, 1, 1, 1, 1]
        sage: [Q.hasse_invariant__OMeara(p) for p in prime_range(20)]                   # needs sage.libs.pari
        [-1, 1, 1, 1, 1, 1, 1, 1]

    ::

        sage: x = polygen(ZZ, 'x')
        sage: K.<a> = NumberField(x^2 - 23)                                             # needs sage.rings.number_field
        sage: Q = DiagonalQuadraticForm(K, [-a, a + 2])                                 # needs sage.rings.number_field
        sage: [Q.hasse_invariant(p) for p in K.primes_above(19)]                        # needs sage.rings.number_field
        [-1, 1]
    """
    # TO DO: Need to deal with the case n=1 separately somewhere!

    Diag = self.rational_diagonal_form()
    R = Diag.base_ring()

    hasse_temp = 1
    n = Diag.dim()

    if R == QQ:
        for j in range(n - 1):
            for k in range(j + 1, n):
                hasse_temp = hasse_temp * hilbert_symbol(Diag[j, j],
                                                         Diag[k, k], p)

    else:
        for j in range(n - 1):
            for k in range(j + 1, n):
                hasse_temp = hasse_temp * R.hilbert_symbol(Diag[j, j],
                                                           Diag[k, k], p)

    return hasse_temp


def hasse_invariant__OMeara(self, p):
    r"""
    Compute the O'Meara Hasse invariant at a prime `p`.

    This is defined on
    p167 of O'Meara's book. If `Q` is diagonal with coefficients `a_i`,
    then the (Cassels) Hasse invariant is given by

    .. MATH::

        c_p = \prod_{i \leq j} (a_i, a_j)_p

    where `(a,b)_p` is the Hilbert symbol at `p`.

    .. WARNING::

        This is different from the (Cassels) Hasse invariant, which
        only allows `i < j` in the product.  That is given by the method
        hasse_invariant(p).

    INPUT:

    - ``p`` -- a prime number > 0 or `-1` for the infinite place

    OUTPUT: `1` or `-1`

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [1,2,3])
        sage: Q.rational_diagonal_form()
        Quadratic form in 2 variables over Rational Field with coefficients:
        [ 1 0 ]
        [ * 2 ]
        sage: [Q.hasse_invariant(p) for p in prime_range(20)]                           # needs sage.libs.pari
        [1, 1, 1, 1, 1, 1, 1, 1]
        sage: [Q.hasse_invariant__OMeara(p) for p in prime_range(20)]                   # needs sage.libs.pari
        [1, 1, 1, 1, 1, 1, 1, 1]

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-1])
        sage: [Q.hasse_invariant(p) for p in prime_range(20)]                           # needs sage.libs.pari
        [1, 1, 1, 1, 1, 1, 1, 1]
        sage: [Q.hasse_invariant__OMeara(p) for p in prime_range(20)]                   # needs sage.libs.pari
        [-1, 1, 1, 1, 1, 1, 1, 1]

    ::

        sage: Q = DiagonalQuadraticForm(ZZ,[1,-1,-1])
        sage: [Q.hasse_invariant(p) for p in prime_range(20)]                           # needs sage.libs.pari
        [-1, 1, 1, 1, 1, 1, 1, 1]
        sage: [Q.hasse_invariant__OMeara(p) for p in prime_range(20)]                   # needs sage.libs.pari
        [-1, 1, 1, 1, 1, 1, 1, 1]

    ::

        sage: x = polygen(ZZ, 'x')
        sage: K.<a> = NumberField(x^2 - 23)                                             # needs sage.rings.number_field
        sage: Q = DiagonalQuadraticForm(K, [-a, a + 2])                                 # needs sage.rings.number_field
        sage: [Q.hasse_invariant__OMeara(p) for p in K.primes_above(19)]                # needs sage.rings.number_field
        [1, 1]
    """
    # TO DO: Need to deal with the case n=1 separately somewhere!

    Diag = self.rational_diagonal_form()
    R = Diag.base_ring()

    hasse_temp = 1
    n = Diag.dim()
    if R == QQ:
        for j in range(n):
            for k in range(j, n):
                hasse_temp = hasse_temp * hilbert_symbol(Diag[j, j],
                                                         Diag[k, k], p)

    else:
        for j in range(n):
            for k in range(j, n):
                hasse_temp = hasse_temp * R.hilbert_symbol(Diag[j, j],
                                                           Diag[k, k], p)

    return hasse_temp


def is_hyperbolic(self, p) -> bool:
    r"""
    Check if the quadratic form is a sum of hyperbolic planes over
    the `p`-adic numbers `\QQ_p` or over the real numbers `\RR`.

    REFERENCES:

    This criterion follows from Cassels's "Rational Quadratic Forms":

    - local invariants for hyperbolic plane (Lemma 2.4, p58)
    - direct sum formulas (Lemma 2.3, p58)

    INPUT:

    - ``p`` -- a prime number > 0 or `-1` for the infinite place

    OUTPUT: boolean

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: Q = DiagonalQuadraticForm(ZZ, [1,1])
        sage: Q.is_hyperbolic(-1)
        False
        sage: Q.is_hyperbolic(2)
        False
        sage: Q.is_hyperbolic(3)
        False
        sage: Q.is_hyperbolic(5)     # Here -1 is a square, so it's true.
        True
        sage: Q.is_hyperbolic(7)
        False
        sage: Q.is_hyperbolic(13)    # Here -1 is a square, so it's true.
        True
    """
    # False for odd-dim'l forms
    if self.dim() % 2:
        return False

    # True for the zero form
    if not self.dim():
        return True

    # Compare local invariants
    # Note: since the dimension is even, the extra powers of 2 in
    #        self.det() := Det(2*Q) don't affect the answer!
    m = ZZ(self.dim() // 2)
    if p == -1:
        return self.signature() == 0

    if p == 2:
        return (QQ(self.det() * (-1) ** m).is_padic_square(p) and
                self.hasse_invariant(p) ==
                (-1) ** m.binomial(2))  # here -1 is hilbert_symbol(-1,-1,2)

    return (QQ(self.det() * (-1) ** m).is_padic_square(p) and
            self.hasse_invariant(p) == 1)


def is_anisotropic(self, p) -> bool:
    r"""
    Check if the quadratic form is anisotropic over the `p`-adic numbers `\QQ_p` or `\RR`.

    INPUT:

    - ``p`` -- a prime number > 0 or `-1` for the infinite place

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1])
        sage: Q.is_anisotropic(2)                                                       # needs sage.libs.pari
        True
        sage: Q.is_anisotropic(3)                                                       # needs sage.libs.pari
        True
        sage: Q.is_anisotropic(5)                                                       # needs sage.libs.pari
        False

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-1])
        sage: Q.is_anisotropic(2)                                                       # needs sage.libs.pari
        False
        sage: Q.is_anisotropic(3)                                                       # needs sage.libs.pari
        False
        sage: Q.is_anisotropic(5)                                                       # needs sage.libs.pari
        False

    ::

        sage: [DiagonalQuadraticForm(ZZ,                                                # needs sage.libs.pari
        ....:                        [1, -least_quadratic_nonresidue(p)]).is_anisotropic(p)
        ....:  for p in prime_range(3, 30)]
        [True, True, True, True, True, True, True, True, True]

    ::

        sage: [DiagonalQuadraticForm(ZZ, [1, -least_quadratic_nonresidue(p),            # needs sage.libs.pari
        ....:                             p, -p*least_quadratic_nonresidue(p)]).is_anisotropic(p)
        ....:  for p in prime_range(3, 30)]
        [True, True, True, True, True, True, True, True, True]
    """
    # TO DO: Should check that p is prime
    if p == -1:
        return self.is_definite()

    n = self.dim()
    D = self.det()

    if n >= 5:
        return False

    if n == 4:
        return (QQ(D).is_padic_square(p) and
                (self.hasse_invariant(p) == - hilbert_symbol(-1, -1, p)))

    if n == 3:
        return self.hasse_invariant(p) != hilbert_symbol(-1, -D, p)

    if n == 2:
        return not QQ(-D).is_padic_square(p)

    if n == 1:
        return self[0, 0] != 0

    raise NotImplementedError("we have not established a convention for 0-dim'l quadratic forms")


def is_isotropic(self, p) -> bool:
    r"""
    Check if `Q` is isotropic over the `p`-adic numbers `\QQ_p` or `\RR`.

    INPUT:

    - ``p`` -- a prime number > 0 or `-1` for the infinite place

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1])
        sage: Q.is_isotropic(2)                                                         # needs sage.libs.pari
        False
        sage: Q.is_isotropic(3)                                                         # needs sage.libs.pari
        False
        sage: Q.is_isotropic(5)                                                         # needs sage.libs.pari
        True

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-1])
        sage: Q.is_isotropic(2)                                                         # needs sage.libs.pari
        True
        sage: Q.is_isotropic(3)                                                         # needs sage.libs.pari
        True
        sage: Q.is_isotropic(5)                                                         # needs sage.libs.pari
        True

    ::

        sage: [DiagonalQuadraticForm(ZZ,                                                # needs sage.libs.pari
        ....:                        [1, -least_quadratic_nonresidue(p)]).is_isotropic(p)
        ....:  for p in prime_range(3, 30)]
        [False, False, False, False, False, False, False, False, False]

    ::

        sage: [DiagonalQuadraticForm(ZZ, [1, -least_quadratic_nonresidue(p),            # needs sage.libs.pari
        ....:                             p, -p*least_quadratic_nonresidue(p)]).is_isotropic(p)
        ....:  for p in prime_range(3, 30)]
        [False, False, False, False, False, False, False, False, False]
    """
    return not self.is_anisotropic(p)


def anisotropic_primes(self):
    """
    Return a list with all of the anisotropic primes of the quadratic form.

    The infinite place is denoted by `-1`.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.anisotropic_primes()                                                    # needs sage.libs.pari
        [2, -1]

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.anisotropic_primes()                                                    # needs sage.libs.pari
        [2, -1]

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1,1])
        sage: Q.anisotropic_primes()                                                    # needs sage.libs.pari
        [-1]
    """
    # Look at all prime divisors of 2 * Det(Q) to find the
    # anisotropic primes...
    possible_primes = prime_divisors(2 * self.det()) + [-1]
    return [p for p in possible_primes if self.is_anisotropic(p)]


def compute_definiteness(self):
    """
    Compute whether the given quadratic form is positive-definite,
    negative-definite, indefinite, degenerate, or the zero form.

    This caches one of the following strings in ``self.__definiteness_string``:
    "pos_def", "neg_def", "indef", "zero", "degenerate".  It is called
    from all routines like: :meth:`is_positive_definite`, :meth:`is_negative_definite`,
    :meth:`is_indefinite`, etc.

    .. NOTE::

        A degenerate form is considered neither definite nor indefinite.

    .. NOTE::

        The zero-dimensional form is considered both positive definite and negative definite.

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1,1])
        sage: Q.compute_definiteness()
        sage: Q.is_positive_definite()
        True
        sage: Q.is_negative_definite()
        False
        sage: Q.is_indefinite()
        False
        sage: Q.is_definite()
        True

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [])
        sage: Q.compute_definiteness()
        sage: Q.is_positive_definite()
        True
        sage: Q.is_negative_definite()
        True
        sage: Q.is_indefinite()
        False
        sage: Q.is_definite()
        True

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,0,-1])
        sage: Q.compute_definiteness()
        sage: Q.is_positive_definite()
        False
        sage: Q.is_negative_definite()
        False
        sage: Q.is_indefinite()
        False
        sage: Q.is_definite()
        False
    """
    # Sanity Check
    from sage.rings.real_mpfr import RR

    if not ((self.base_ring() == ZZ) or (self.base_ring() == QQ) or (self.base_ring() == RR)):
        raise NotImplementedError("we can only check definiteness over ZZ, QQ, and RR for now")

    # Some useful variables
    n = self.dim()

    # Deal with the zero-diml form
    if n == 0:
        self.__definiteness_string = "zero"
        return

    sig_pos, sig_neg, sig_zer = self.signature_vector()

    # Determine and cache the definiteness string
    if sig_zer > 0:
        self.__definiteness_string = "degenerate"
        return
    if sig_neg == n:
        self.__definiteness_string = "neg_def"
        return
    if sig_pos == n:
        self.__definiteness_string = "pos_def"
        return
    self.__definiteness_string = "indefinite"
    return


def compute_definiteness_string_by_determinants(self):
    """
    Compute the (positive) definiteness of a quadratic form by looking
    at the signs of all of its upper-left subdeterminants.  See also
    :meth:`compute_definiteness` for more documentation.

    OUTPUT: string describing the definiteness

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1,1])
        sage: Q.compute_definiteness_string_by_determinants()
        'pos_def'

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [])
        sage: Q.compute_definiteness_string_by_determinants()
        'zero'

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,0,-1])
        sage: Q.compute_definiteness_string_by_determinants()
        'degenerate'

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-1])
        sage: Q.compute_definiteness_string_by_determinants()
        'indefinite'

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [-1,-1])
        sage: Q.compute_definiteness_string_by_determinants()
        'neg_def'
    """
    # Sanity Check
    from sage.rings.real_mpfr import RR

    if not ((self.base_ring() == ZZ) or (self.base_ring() == QQ) or (self.base_ring() == RR)):
        raise NotImplementedError("we can only check definiteness over ZZ, QQ, and RR for now")

    from sage.functions.generalized import sgn

    # Some useful variables
    n = self.dim()
    M = self.matrix()

    # Deal with the zero-diml form
    if n == 0:
        return "zero"

    # Deal with degenerate forms
    if self.det() == 0:
        return "degenerate"

    # Check the sign of the ratios of consecutive determinants of the upper triangular r x r submatrices
    first_coeff = self[0, 0]
    for r in range(1, n + 1):
        I = list(range(r))
        new_det = M.matrix_from_rows_and_columns(I, I).det()

        # Check for a (non-degenerate) zero -- so it's indefinite
        if new_det == 0:
            return "indefinite"

        # Check for a change of signs in the upper r x r submatrix -- so it's indefinite
        if sgn(first_coeff)**r != sgn(new_det):
            return "indefinite"

    # Here all ratios of determinants have the correct sign, so the matrix is (pos or neg) definite.
    return "pos_def" if first_coeff > 0 else "neg_def"


def is_positive_definite(self) -> bool:
    """
    Determine if the given quadratic form is positive-definite.

    .. NOTE::

        A degenerate form is considered neither definite nor indefinite.

    .. NOTE::

        The zero-dimensional form is considered both positive definite and negative definite.

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5])
        sage: Q.is_positive_definite()
        True

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-3,5])
        sage: Q.is_positive_definite()
        False
    """
    # Try to use the cached value
    try:
        def_str = self.__definiteness_string
    except AttributeError:
        self.compute_definiteness()
        def_str = self.__definiteness_string

    # Return the answer
    return (def_str == "pos_def") or (def_str == "zero")


def is_negative_definite(self) -> bool:
    """
    Determine if the given quadratic form is negative-definite.

    .. NOTE::

        A degenerate form is considered neither definite nor indefinite.

    .. NOTE::

        The zero-dimensional form is considered both positive definite and negative definite.

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [-1,-3,-5])
        sage: Q.is_negative_definite()
        True

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-3,5])
        sage: Q.is_negative_definite()
        False
    """
    # Try to use the cached value
    try:
        def_str = self.__definiteness_string
    except AttributeError:
        self.compute_definiteness()
        def_str = self.__definiteness_string

    # Return the answer
    return (def_str == "neg_def") or (def_str == "zero")


def is_indefinite(self) -> bool:
    """
    Determine if the given quadratic form is indefinite.

    .. NOTE::

        A degenerate form is considered neither definite nor indefinite.

    .. NOTE::

        The zero-dimensional form is not considered indefinite.

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [-1,-3,-5])
        sage: Q.is_indefinite()
        False

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-3,5])
        sage: Q.is_indefinite()
        True
    """
    # Try to use the cached value
    try:
        def_str = self.__definiteness_string
    except AttributeError:
        self.compute_definiteness()
        def_str = self.__definiteness_string

    # Return the answer
    return def_str == "indefinite"


def is_definite(self) -> bool:
    """
    Determine if the given quadratic form is (positive or negative) definite.

    .. NOTE::

        A degenerate form is considered neither definite nor indefinite.

    .. NOTE::

        The zero-dimensional form is considered indefinite.

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [-1,-3,-5])
        sage: Q.is_definite()
        True

        sage: Q = DiagonalQuadraticForm(ZZ, [1,-3,5])
        sage: Q.is_definite()
        False
    """
    # Try to use the cached value
    try:
        def_str = self.__definiteness_string
    except AttributeError:
        self.compute_definiteness()
        def_str = self.__definiteness_string

    # Return the answer
    return (def_str == "pos_def") or (def_str == "neg_def") or (def_str == "zero")



# distutils: libraries = NTL_LIBRARIES gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++
# sage.doctest: needs sage.rings.padics
r"""
`p`-adic Extension Element

A common superclass for all elements of extension rings and field of `\ZZ_p` and
`\QQ_p`.

AUTHORS:

- David Roe (2007): initial version

- Julian Rueth (2012-10-18): added residue
"""
# ****************************************************************************
#       Copyright (C) 2007-2010 David Roe <roed.math@gmail.com>
#                     2012 Julian Rueth <julian.rueth@fsfe.org>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.padics.pow_computer cimport PowComputer_class
from sage.libs.ntl.ntl_ZZ_p cimport ntl_ZZ_p

cdef class pAdicExtElement(pAdicGenericElement):
    cdef int _set_from_list(self, L) except -1:
        """
        Set ``self`` from a list.

        The list should either be uniform in type, or all of the entries
        should be coercible to integers. If any of the entries in ``L``
        is a list, ``L`` will be cast to a ZZ_pEX.

        INPUT:

        - ``L`` -- list
        """
        raise NotImplementedError

    cdef int _set_from_list_rel(self, L, long relprec) except -1:
        raise NotImplementedError

    cdef int _set_from_list_abs(self, L, long absprec) except -1:
        raise NotImplementedError

    cdef int _set_from_list_both(self, L, long absprec, long relprec) except -1:
        raise NotImplementedError

    cdef int _set_from_ZZX(self, ZZX_c poly) except -1:
        """
        Set from a ZZX_c, choosing how to handle based on the
        precision type of ``self.parent()``.

        Fixed modulus elements should override this function.

        This function is not used internally.
        """
        if self.parent().is_capped_relative():
            self._set_from_ZZX_rel(poly, (<PowComputer_class>self.parent().prime_pow).prec_cap)
        elif self.parent().is_capped_absolute():
            self._set_from_ZZX_abs(poly, (<PowComputer_class>self.parent().prime_pow).prec_cap)
        else:
            raise RuntimeError("_set_from_ZZX should have been overridden")

    cdef int _set_from_ZZX_rel(self, ZZX_c poly, long relprec) except -1:
        """
        Set from a ZZX_c with bounded relative precision.

        Capped relative elements should override this function, so the
        default implementation is for capped absolute.

        This function is not used internally.
        """
        self._set_from_ZZX_both(poly, (<PowComputer_class>self.parent().prime_pow).prec_cap, relprec)

    cdef int _set_from_ZZX_abs(self, ZZX_c poly, long absprec) except -1:
        """
        Set from a ZZX_c with bounded absolute precision.

        Capped absolute elements should override this function, so the
        default implementation is for capped relative.

        This function is not used internally.
        """
        self._set_from_ZZX_both(poly, absprec, (<PowComputer_class>self.parent().prime_pow).prec_cap)

    cdef int _set_from_ZZX_both(self, ZZX_c poly, long absprec, long relprec) except -1:
        """
        Set from a ZZX_c with both absolute and relative precisions bounded.

        This function should be overridden for both capped absolute
        and capped relative elements.

        This function is not used internally.
        """
        if self.parent().is_fixed_mod():
            self._set_from_ZZX(poly)
        else:
            raise RuntimeError("_set_from_ZZX_both should have been overridden")

    cdef int _set_from_ZZ_pX(self, ZZ_pX_c* poly, ntl_ZZ_pContext_class ctx) except -1:
        """
        Set ``self`` from a ZZ_pX defined with context ctx.

        This function should be overridden for fixed modulus elements.

        This function is not used internally.
        """
        if self.parent().is_capped_relative():
            self._set_from_ZZ_pX_rel(poly, ctx, (<PowComputer_class>self.parent().prime_pow).prec_cap)
        elif self.parent().is_capped_absolute():
            self._set_from_ZZ_pX_abs(poly, ctx, (<PowComputer_class>self.parent().prime_pow).prec_cap)
        else:
            raise RuntimeError("_set_from_ZZ_pX should have been overridden")

    cdef int _set_from_ZZ_pX_rel(self, ZZ_pX_c* poly, ntl_ZZ_pContext_class ctx, long relprec) except -1:
        """
        Set from a ZZ_pX_c with bounded relative precision.

        Capped relative rings should override this function, so the
        default implementation is for capped absolute.

        This function is not used internally.
        """
        self._set_from_ZZ_pX_both(poly, ctx, (<PowComputer_class>self.parent().prime_pow).prec_cap, relprec)

    cdef int _set_from_ZZ_pX_abs(self, ZZ_pX_c* poly, ntl_ZZ_pContext_class ctx, long absprec) except -1:
        """
        Set from a ZZ_pX_c with bounded absolute precision.

        Capped absolute rings should override this function, so the
        default implementation is for capped relative.

        This function is not used internally.
        """
        self._set_from_ZZ_pX_both(poly, ctx, absprec, (<PowComputer_class>self.parent().prime_pow).prec_cap)

    cdef int _set_from_ZZ_pX_both(self, ZZ_pX_c* poly, ntl_ZZ_pContext_class ctx, long absprec, long relprec) except -1:
        """
        Set from a ZZ_pX_c with both absolute and relative precision bounded.

        This function should be overridden by both capped absolute and capped relative elements.

        This function is not used internally.
        """
        if self.parent().is_fixed_mod():
            self._set_from_ZZ_pX(poly, ctx)
        else:
            raise RuntimeError("_set_from_ZZ_pX_both should have been overridden")

    cdef int _set_from_ZZ_pE(self, ZZ_pE_c* poly, ntl_ZZ_pEContext_class ctx) except -1:
        """
        Set from a ZZ_pE_c.

        This function is not used internally.
        """
        if self.parent().is_capped_relative():
            self._set_from_ZZ_pE_rel(poly, ctx, (<PowComputer_class>self.parent().prime_pow).prec_cap)
        elif self.parent().is_capped_absolute():
            self._set_from_ZZ_pE_abs(poly, ctx, (<PowComputer_class>self.parent().prime_pow).prec_cap)
        else:
            raise RuntimeError("_set_from_ZZ_pE should have been overridden")

    cdef int _set_from_ZZ_pE_rel(self, ZZ_pE_c* poly, ntl_ZZ_pEContext_class ctx, long relprec) except -1:
        """
        Set from a ZZ_pE_c with bounded relative precision.

        Capped relative rings should override this function, so the
        default implementation is for capped absolute.

        This function is not used internally.
        """
        self._set_from_ZZ_pE_both(poly, ctx, (<PowComputer_class>self.parent().prime_pow).prec_cap, relprec)

    cdef int _set_from_ZZ_pE_abs(self, ZZ_pE_c* poly, ntl_ZZ_pEContext_class ctx, long absprec) except -1:
        """
        Set from a ZZ_pE_c with bounded absolute precision.

        Capped absolute elements should override this function, so the
        default implementation is for capped relative.

        This function is not used internally.
        """
        self._set_from_ZZ_pE_both(poly, ctx, absprec, (<PowComputer_class>self.parent().prime_pow).prec_cap)

    cdef int _set_from_ZZ_pE_both(self, ZZ_pE_c* poly, ntl_ZZ_pEContext_class ctx, long absprec, long relprec) except -1:
        """
        Set from a ZZ_pE_c with both absolute and relative precision bounded.

        Capped absolute and capped relative elements should override
        this function.

        This function is not used internally.
        """
        if self.parent().is_fixed_mod():
            self._set_from_ZZ_pE(poly, ctx)
        else:
            raise RuntimeError("_set_from_ZZ_pE_both should have been overridden")

    cdef int _set_from_ZZ_pEX(self, ZZ_pEX_c* poly, ntl_ZZ_pEContext_class ctx) except -1:
        """
        Set ``self`` from a ZZ_pEX_c.

        Fixed modulus elements should override this function.

        This function is not used internally.
        """
        if self.parent().is_capped_relative():
            self._set_from_ZZ_pEX_rel(poly, ctx, (<PowComputer_class>self.parent().prime_pow).prec_cap)
        elif self.parent().is_capped_absolute():
            self._set_from_ZZ_pEX_abs(poly, ctx, (<PowComputer_class>self.parent().prime_pow).prec_cap)
        else:
            raise RuntimeError("_set_from_ZZ_pEX should have been overridden")

    cdef int _set_from_ZZ_pEX_rel(self, ZZ_pEX_c* poly, ntl_ZZ_pEContext_class ctx, long relprec) except -1:
        """
        Set from a ZZ_pEX_c with bounded relative precision.

        Capped relative elements should override this function, so the
        default implementation is for capped absolute.

        This function is not used internally.
        """
        self._set_from_ZZ_pEX_both(poly, ctx, (<PowComputer_class>self.parent().prime_pow).prec_cap, relprec)

    cdef int _set_from_ZZ_pEX_abs(self, ZZ_pEX_c* poly, ntl_ZZ_pEContext_class ctx, long absprec) except -1:
        """
        Set from a ZZ_pEX_c with bounded absolute precision.

        Capped absolute elements should override this function, so the
        default implementation is for capped relative.

        This function is not used internally.
        """
        self._set_from_ZZ_pEX_both(poly, ctx, absprec, (<PowComputer_class>self.parent().prime_pow).prec_cap)

    cdef int _set_from_ZZ_pEX_both(self, ZZ_pEX_c* poly, ntl_ZZ_pEContext_class ctx, long absprec, long relprec) except -1:
        """
        Set from a ZZ_pEX_c with both absolute and relative precision bounded.

        Capped absolute and capped relative elements should override
        this function.

        This function is not used internally.
        """
        if self.parent().is_fixed_mod():
            self._set_from_ZZ_pEX(poly, ctx)
        else:
            raise RuntimeError("_set_from_ZZ_pEX_both should have been overridden")

    cdef long _check_ZZ_pContext(self, ntl_ZZ_pContext_class ctx) except -1:
        raise NotImplementedError

    cdef long _check_ZZ_pEContext(self, ntl_ZZ_pEContext_class ctx) except -1:
        raise NotImplementedError

    cdef ext_p_list(self, bint pos):
        raise NotImplementedError

    cdef ext_p_list_precs(self, bint pos, long prec):
        raise NotImplementedError

    def _const_term_test(self):
        """
        Return the constant term of a polynomial representing ``self``.

        This function is mainly for troubleshooting, and the meaning
        of the return value will depend on whether ``self`` is capped
        relative or otherwise.

        EXAMPLES::

            sage: R = Zp(5,5)
            sage: S.<x> = R[]
            sage: f = x^5 + 75*x^3 - 15*x^2 + 125*x - 5
            sage: W.<w> = R.ext(f)
            sage: a = W(566)
            sage: a._const_term_test()
            566
        """
        cdef ntl_ZZ_p ans = ntl_ZZ_p(modulus=self.parent().prime())
        ans.x = self._const_term()
        return ans

    cdef ZZ_p_c _const_term(self) noexcept:
        raise NotImplementedError

    def _ext_p_list(self, pos):
        r"""
        Return a list of integers (in the Eisenstein case) or a list
        of lists of integers (in the unramified case).

        ``self`` can be reconstructed as a sum of elements of the list times
        powers of the uniformiser (in the Eisenstein case), or as a sum of
        powers of the p times polynomials in the generator (in the
        unramified case).

        Note that zeros are truncated from the returned list, so you must
        use the :func:`valuation()` function to completely recover ``self``.

        INPUT:

        - ``pos`` -- boolean; if ``True``, all integers will be in the range
          `[0,p-1]`, otherwise they will be in the range `[(1-p)/2, p/2]`

        OUTPUT:

        - a list of integers or list of lists giving the
          series expansion of ``self``

        EXAMPLES::

            sage: R = Zp(5,5)
            sage: S.<x> = R[]
            sage: f = x^5 + 75*x^3 - 15*x^2 + 125*x - 5
            sage: W.<w> = R.ext(f)
            sage: y = W(775, 19); y
            w^10 + 4*w^12 + 2*w^14 + w^15 + 2*w^16 + 4*w^17 + w^18 + O(w^19)
            sage: y._ext_p_list(True)
            [1, 0, 4, 0, 2, 1, 2, 4, 1]
            sage: y._ext_p_list(False)
            [1, 0, -1, 0, 2, 1, 2, 0, 1]
        """
        return self.ext_p_list(pos)

    def frobenius(self, arithmetic=True):
        r"""
        Return the image of this element under the Frobenius automorphism
        applied to its parent.

        INPUT:

        - ``arithmetic`` -- whether to apply the arithmetic Frobenius (acting
          by raising to the `p`-th power on the residue field). If ``False`` is
          provided, the image of geometric Frobenius (raising to the `(1/p)`-th
          power on the residue field) will be returned instead.

        EXAMPLES::

            sage: R.<a> = Zq(5^4,3)
            sage: a.frobenius()
            (a^3 + a^2 + 3*a) + (3*a + 1)*5 + (2*a^3 + 2*a^2 + 2*a)*5^2 + O(5^3)
            sage: f = R.defining_polynomial()
            sage: f(a)
            O(5^3)
            sage: f(a.frobenius())
            O(5^3)
            sage: for i in range(4): a = a.frobenius()
            sage: a
            a + O(5^3)

            sage: K.<a> = Qq(7^3,4)
            sage: b = (a+1)/7
            sage: c = b.frobenius(); c
            (3*a^2 + 5*a + 1)*7^-1 + (6*a^2 + 6*a + 6) + (4*a^2 + 3*a + 4)*7 + (6*a^2 + a + 6)*7^2 + O(7^3)
            sage: c.frobenius().frobenius()
            (a + 1)*7^-1 + O(7^3)

        An error will be raised if the parent of ``self`` is a ramified extension::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = Qp(5).extension(x^2 - 5)
            sage: a.frobenius()
            Traceback (most recent call last):
            ...
            NotImplementedError: Frobenius automorphism only implemented for unramified extensions
        """
        R = self.parent()
        if R.absolute_e() != 1:
            raise NotImplementedError("Frobenius automorphism only implemented for unramified extensions")
        if self.is_zero():
            return self
        L = self.teichmuller_expansion()
        ppow = R.uniformizer_pow(self.valuation())
        if arithmetic:
            exp = R.prime()
        else:
            exp = R.prime()**(R.absolute_degree()-1)
        ans = ppow * L[0]**exp
        for m in range(1, len(L)):
            ppow = ppow << 1
            ans += ppow * L[m]**exp
        return ans

    cpdef bint _is_base_elt(self, p) except -1:
        r"""
        Return ``True`` if this element is an element of Zp or Qp (rather than
        an extension).

        INPUT:

        - ``p`` -- a prime, which is compared with the parent of this element

        EXAMPLES::

            sage: K.<a> = Qq(7^3,4)
            sage: a._is_base_elt(5)
            False
        """
        return False

    def residue(self, absprec=1, field=None, check_prec=True):
        r"""
        Reduces this element modulo `\pi^\mathrm{absprec}`.

        INPUT:

        - ``absprec`` -- nonnegative integer (default: 1)

        - ``field`` -- boolean (default: ``None``); for precision 1, whether to return
          an element of the residue field or a residue ring.  Currently unused.

        - ``check_prec`` -- boolean (default: ``True``); whether to raise an error if this
          element has insufficient precision to determine the reduction.  Errors are never
          raised for fixed-mod or floating-point types.

        OUTPUT: this element reduced modulo `\pi^\mathrm{absprec}`

        If ``absprec`` is zero, then as an element of `\ZZ/(1)`.

        If ``absprec`` is one, then as an element of the residue field.

        .. NOTE::

            Only implemented for ``absprec`` less than or equal to one.

        AUTHORS:

        - Julian Rueth (2012-10-18): initial version

        EXAMPLES:

        Unramified case::

            sage: R = ZpCA(3,5)
            sage: S.<a> = R[]
            sage: W.<a> = R.extension(a^2 + 9*a + 1)
            sage: (a + 1).residue(1)
            a0 + 1
            sage: a.residue(2)
            Traceback (most recent call last):
            ...
            NotImplementedError: reduction modulo p^n with n>1

        Eisenstein case::

            sage: R = ZpCA(3,5)
            sage: S.<a> = R[]
            sage: W.<a> = R.extension(a^2 + 9*a + 3)
            sage: (a + 1).residue(1)
            1
            sage: a.residue(2)
            Traceback (most recent call last):
            ...
            NotImplementedError: residue() not implemented in extensions for absprec larger than one

        TESTS::

            sage: K = Qp(3,5)
            sage: S.<a> = R[]
            sage: W.<a> = R.extension(a^2 + 9*a + 1)
            sage: (a/3).residue(0)
            Traceback (most recent call last):
            ...
            ValueError: element must have nonnegative valuation in order to compute residue

            sage: R = ZpFM(3,5)
            sage: S.<a> = R[]
            sage: W.<a> = R.extension(a^2 + 3)
            sage: W.one().residue(0)
            0
            sage: a.residue(-1)
            Traceback (most recent call last):
            ...
            ValueError: cannot reduce modulo a negative power of the uniformizer
            sage: a.residue(16)
            Traceback (most recent call last):
            ...
            NotImplementedError: residue() not implemented in extensions for absprec larger than one
        """
        if absprec < 0:
            raise ValueError("cannot reduce modulo a negative power of the uniformizer")
        if self.valuation() < 0:
            raise ValueError("element must have nonnegative valuation in order to compute residue")
        R = self.parent()
        if check_prec and (R.is_fixed_mod() or R.is_floating_point()):
            check_prec = False
        if check_prec and absprec > self.precision_absolute():
            from precision_error import PrecisionError
            raise PrecisionError("not enough precision known in order to compute residue")
        if field and absprec != 1:
            raise ValueError("field keyword may only be set at precision 1")

        if absprec == 0:
            from sage.rings.finite_rings.integer_mod import Mod
            return Mod(0, 1)
        elif absprec == 1:
            return R.residue_field()(self.expansion(0))
        else:
            raise NotImplementedError("residue() not implemented in extensions for absprec larger than one")



# sage.doctest: needs sage.libs.pari
"""
Submodules of spaces of modular forms

EXAMPLES::

    sage: M = ModularForms(Gamma1(13),2); M
    Modular Forms space of dimension 13 for
     Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
    sage: M.eisenstein_subspace()
    Eisenstein subspace of dimension 11 of Modular Forms space of dimension 13 for
     Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
    sage: M == loads(dumps(M))
    True
    sage: M.cuspidal_subspace()
    Cuspidal subspace of dimension 2 of Modular Forms space of dimension 13 for
     Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
"""

#########################################################################
#       Copyright (C) 2004--2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#########################################################################

from .space import ModularFormsSpace

import sage.modular.hecke.submodule


class ModularFormsSubmodule(ModularFormsSpace,
                            sage.modular.hecke.submodule.HeckeSubmodule):
    """
    A submodule of an ambient space of modular forms.
    """
    def __init__(self, ambient_module, submodule, dual=None, check=False):
        """
        INPUT:

        - ``ambient_module`` -- ModularFormsSpace
        - ``submodule`` -- a submodule of the ambient space
        - ``dual_module`` -- (default: ``None``) ignored
        - ``check`` -- boolean (default: ``False``); whether to check that the
          submodule is Hecke equivariant

        EXAMPLES::

          sage: M = ModularForms(Gamma1(13),2); M
          Modular Forms space of dimension 13 for Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
          sage: M.eisenstein_subspace()
          Eisenstein subspace of dimension 11 of Modular Forms space of dimension 13 for Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
        """
        A = ambient_module
        sage.modular.hecke.submodule.HeckeSubmodule.__init__(self, A, submodule, check=check)
        ModularFormsSpace.__init__(self, A.group(), A.weight(),
                                         A.character(), A.base_ring())

    def _repr_(self):
        """
        EXAMPLES::

          sage: ModularForms(Gamma1(13),2).eisenstein_subspace()._repr_()
          'Eisenstein subspace of dimension 11 of Modular Forms space of dimension 13 for Congruence Subgroup Gamma1(13) of weight 2 over Rational Field'
        """
        return "Modular Forms subspace of dimension %s of %s" % (self.dimension(), self.ambient_module())

    def _compute_coefficients(self, element, X):
        """
        Compute all coefficients of the modular form element in ``self`` for
        indices in X.

        TODO: Implement this function.

        EXAMPLES::

            sage: M = ModularForms(6,4).cuspidal_subspace()
            sage: M._compute_coefficients( M.basis()[0], range(1,100) )
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def _compute_q_expansion_basis(self, prec):
        """
        Compute ``q_expansions`` to precision ``prec`` for each element in
        ``self.basis()``.

        EXAMPLES::

            sage: M = ModularForms(Gamma1(13),2); M
            Modular Forms space of dimension 13 for Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
            sage: S = M.eisenstein_subspace(); S
            Eisenstein subspace of dimension 11 of Modular Forms space of dimension 13 for Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
            sage: S._compute_q_expansion_basis(5)
            [1 + O(q^5),
             q + O(q^5),
             q^2 + O(q^5),
             q^3 + O(q^5),
             q^4 + O(q^5),
             O(q^5),
             O(q^5),
             O(q^5),
             O(q^5),
             O(q^5),
             O(q^5)]
        """
        A = self.ambient_module()
        return [A._q_expansion(element=f.element(), prec=prec)
                for f in self.basis()]


# TODO
class ModularFormsSubmoduleWithBasis(ModularFormsSubmodule):
    pass



r"""
Composition tableaux

AUTHORS:

- Chris Berg, Jeff Ferreira (2012-9): initial version
"""
from sage.sets.disjoint_union_enumerated_sets import DisjointUnionEnumeratedSets
from sage.sets.non_negative_integers import NonNegativeIntegers
from sage.sets.family import Family
from sage.misc.classcall_metaclass import ClasscallMetaclass
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.combinat.composition import Composition, Compositions
from sage.combinat.partition import Partition
from sage.combinat.combinat import CombinatorialElement
from sage.rings.integer import Integer
from sage.combinat.backtrack import GenericBacktracker
import copy


class CompositionTableau(CombinatorialElement, metaclass=ClasscallMetaclass):
    r"""
    A composition tableau.

    A *composition tableau* `t` of shape `I = (I_1, \ldots, I_{\ell})` is an
    array of boxes in rows,  `I_i` boxes in row `i`, filled with positive
    integers such that:

    1) the entries in the rows of `t` weakly decrease from left to right,
    2) the left-most column of `t` strictly increase from top to bottom.
    3) Add zero entries to the rows of `t` until the resulting array is
       rectangular of  shape `\ell \times m`. For `1 \leq i < j \leq \ell,
       2 \leq k \leq m` and `(t(j,k) \neq 0`, and also if `t(j,k) \geq t(i,k))`
       implies `t(j,k) > t(i,k-1).`

    INPUT:

    - ``t`` -- list of lists

    EXAMPLES::

        sage: CompositionTableau([[1],[2,2]])
        [[1], [2, 2]]
        sage: CompositionTableau([[1],[3,2],[4,4]])
        [[1], [3, 2], [4, 4]]
        sage: CompositionTableau([])
        []
    """
    @staticmethod
    def __classcall_private__(self, t):
        r"""
        This ensures that a composition tableau is only ever constructed as
        an ``element_class`` call of an appropriate parent.

        TESTS::

            sage: t = CompositionTableau([[1],[2,2]])
            sage: TestSuite(t).run()

            sage: t.parent()
            Composition Tableaux
            sage: t.category()
            Category of elements of Composition Tableaux
        """
        if isinstance(t, CompositionTableau):
            return t
        return CompositionTableaux_all().element_class(CompositionTableaux_all(), t)

    def __init__(self, parent, t):
        r"""
        Initialize a composition tableau.

        TESTS::

            sage: t = CompositionTableaux()([[1],[2,2]])
            sage: s = CompositionTableaux(3)([[1],[2,2]])
            sage: s == t
            True
            sage: t.parent()
            Composition Tableaux
            sage: s.parent()
            Composition Tableaux of size 3 and maximum entry 3
            sage: r = CompositionTableaux()(s)
            sage: r.parent()
            Composition Tableaux
        """
        if isinstance(t, CompositionTableau):
            CombinatorialElement.__init__(self, parent, t._list)
            return

        # CombinatorialObject verifies that t is a list
        # We must verify t is a list of lists
        if not all(isinstance(row, list) for row in t):
            raise ValueError("a composition tableau must be a list of lists")

        if any(not r for r in t):
            raise ValueError("a composition tableau must be a list of non-empty lists")

        # Verify rows weakly decrease from left to right
        for row in t:
            if any(row[i] < row[i + 1] for i in range(len(row) - 1)):
                raise ValueError("rows must weakly decrease from left to right")

        # Verify leftmost column strictly increases from top to bottom
        first_col = [row[0] for row in t if t != [[]]]
        if any(first_col[i] >= first_col[i + 1] for i in range(len(t) - 1)):
            raise ValueError("leftmost column must strictly increase from top to bottom")

        # Verify triple condition
        l = len(t)
        m = max((len(r) for r in t), default=0)
        TT = [row+[0]*(m-len(row)) for row in t]
        for i in range(l):
            for j in range(i+1, l):
                for k in range(1, m):
                    if TT[j][k] and TT[i][k] <= TT[j][k] <= TT[i][k-1]:
                        raise ValueError("triple condition must be satisfied")

        CombinatorialElement.__init__(self, parent, t)

    def _repr_diagram(self) -> str:
        r"""
        Return a string representation of ``self`` as an array.

        EXAMPLES::

            sage: t = CompositionTableau([[1],[3,2],[4,4]])
            sage: print(t._repr_diagram())
              1
              3  2
              4  4
        """
        return '\n'.join("".join("%3s" % str(x) for x in row)
                         for row in self)

    def __call__(self, *cell):
        r"""
        Return the value in the corresponding cell of ``self``.

        EXAMPLES::

            sage: t = CompositionTableau([[1],[3,2],[4,4]])
            sage: t(1,1)
            2
            sage: t(2,0)
            4
            sage: t(2,2)
            Traceback (most recent call last):
            ...
            IndexError: the cell (2,2) is not contained in [[1], [3, 2], [4, 4]]
        """
        try:
            i, j = cell
        except ValueError:
            i, j = cell[0]

        try:
            return self[i][j]
        except IndexError:
            raise IndexError("the cell (%d,%d) is not contained in %s" % (i, j, self))

    def pp(self):
        r"""
        Return a pretty print string of ``self``.

        EXAMPLES::

            sage: CompositionTableau([[1],[3,2],[4,4]]).pp()
            1
            3  2
            4  4
        """
        print(self._repr_diagram())

    def size(self):
        r"""
        Return the number of boxes in ``self``.

        EXAMPLES::

            sage: CompositionTableau([[1],[3,2],[4,4]]).size()
            5
        """
        return sum(len(row) for row in self)

    def weight(self):
        r"""
        Return a composition where entry `i` is the number of times that `i` appears in
        ``self``.

        EXAMPLES::

            sage: CompositionTableau([[1],[3,2],[4,4]]).weight()
            [1, 1, 1, 2, 0]
        """
        w = {i: 0 for i in range(1, self.size() + 1)}
        for row in self:
            for i in row:
                w[i] += 1
        return Composition([w[i] for i in range(1, self.size() + 1)])

    def descent_set(self):
        r"""
        Return the set of all `i` that do *not* have `i+1` appearing strictly
        to the left of `i` in ``self``.

        EXAMPLES::

            sage: CompositionTableau([[1],[3,2],[4,4]]).descent_set()
            [1, 3]
        """
        cols = {i: col for row in self for col, i in enumerate(row)}
        return sorted(i for i in cols if i + 1 in cols and cols[i + 1] >= cols[i])

    def descent_composition(self):
        r"""
        Return the composition corresponding to the set of all `i` that do
        not have `i+1` appearing strictly to the left of `i` in ``self``.

        EXAMPLES::

            sage: CompositionTableau([[1],[3,2],[4,4]]).descent_composition()
            [1, 2, 2]
        """
        return Composition(from_subset=(self.descent_set(), self.size()))

    def shape_composition(self):
        r"""
        Return a Composition object which is the shape of ``self``.

        EXAMPLES::

            sage: CompositionTableau([[1,1],[3,2],[4,4,3]]).shape_composition()
            [2, 2, 3]
            sage: CompositionTableau([[2,1],[3],[4]]).shape_composition()
            [2, 1, 1]
        """
        return Composition([len(row) for row in self])

    def shape_partition(self):
        r"""
        Return a partition which is the shape of ``self`` sorted into weakly
        decreasing order.

        EXAMPLES::

            sage: CompositionTableau([[1,1],[3,2],[4,4,3]]).shape_partition()
            [3, 2, 2]
            sage: CompositionTableau([[2,1],[3],[4]]).shape_partition()
            [2, 1, 1]
        """
        return Partition(sorted((len(row) for row in self), reverse=True))

    def is_standard(self) -> bool:
        r"""
        Return ``True`` if ``self`` is a standard composition tableau and
        ``False`` otherwise.

        EXAMPLES::

            sage: CompositionTableau([[1,1],[3,2],[4,4,3]]).is_standard()
            False
            sage: CompositionTableau([[2,1],[3],[4]]).is_standard()
            True
        """
        entries = sum(self, [])
        return sorted(entries) == list(range(1, self.size() + 1))


class CompositionTableaux(UniqueRepresentation, Parent):
    r"""
    Composition tableaux.

    INPUT:

    - ``size`` -- the size of the composition tableaux
    - ``shape`` -- the shape of the composition tableaux
    - ``max_entry`` -- the maximum entry for the composition tableaux

    The first argument is interpreted as ``size`` or ``shape`` depending on
    whether it is an integer or a composition.

    EXAMPLES::

        sage: CT = CompositionTableaux(3); CT
        Composition Tableaux of size 3 and maximum entry 3
        sage: list(CT)
        [[[1], [2], [3]],
         [[1], [2, 2]],
         [[1], [3, 2]],
         [[1], [3, 3]],
         [[2], [3, 3]],
         [[1, 1], [2]],
         [[1, 1], [3]],
         [[2, 1], [3]],
         [[2, 2], [3]],
         [[1, 1, 1]],
         [[2, 1, 1]],
         [[2, 2, 1]],
         [[2, 2, 2]],
         [[3, 1, 1]],
         [[3, 2, 1]],
         [[3, 2, 2]],
         [[3, 3, 1]],
         [[3, 3, 2]],
         [[3, 3, 3]]]

        sage: CT = CompositionTableaux([1,2,1]); CT
        Composition tableaux of shape [1, 2, 1] and maximum entry 4
        sage: list(CT)
        [[[1], [2, 2], [3]],
         [[1], [2, 2], [4]],
         [[1], [3, 2], [4]],
         [[1], [3, 3], [4]],
         [[2], [3, 3], [4]]]

        sage: CT = CompositionTableaux(shape=[1,2,1],max_entry=3); CT
        Composition tableaux of shape [1, 2, 1] and maximum entry 3
        sage: list(CT)
        [[[1], [2, 2], [3]]]

        sage: CT = CompositionTableaux(2,max_entry=3); CT
        Composition Tableaux of size 2 and maximum entry 3
        sage: list(CT)
        [[[1], [2]],
         [[1], [3]],
         [[2], [3]],
         [[1, 1]],
         [[2, 1]],
         [[2, 2]],
         [[3, 1]],
         [[3, 2]],
         [[3, 3]]]

        sage: CT = CompositionTableaux(0); CT
        Composition Tableaux of size 0 and maximum entry 0
        sage: list(CT)
        [[]]
    """
    @staticmethod
    def __classcall_private__(cls, *args, **kwargs):
        r"""
        This is a factory class which returns the appropriate parent based on
        arguments.  See the documentation for :class:`CompositionTableaux` for
        more information.

        TESTS::

            sage: CT = CompositionTableaux(3); CT
            Composition Tableaux of size 3 and maximum entry 3
            sage: CT = CompositionTableaux(size=3); CT
            Composition Tableaux of size 3 and maximum entry 3
            sage: CT = CompositionTableaux([1,2]); CT
            Composition tableaux of shape [1, 2] and maximum entry 3
            sage: CT = CompositionTableaux(shape=[1,2]); CT
            Composition tableaux of shape [1, 2] and maximum entry 3
            sage: CT = CompositionTableaux(shape=[]); CT
            Composition tableaux of shape [] and maximum entry 0
            sage: CT = CompositionTableaux(0); CT
            Composition Tableaux of size 0 and maximum entry 0
            sage: CT = CompositionTableaux(max_entry=3); CT
            Composition tableaux with maximum entry 3
            sage: CT = CompositionTableaux([1,2],max_entry=3); CT
            Composition tableaux of shape [1, 2] and maximum entry 3
            sage: CT = CompositionTableaux(size=2,shape=[1,2]); CT
            Traceback (most recent call last):
            ...
            ValueError: size and shape are different sizes
        """
        # Process keyword arguments first
        n = kwargs.get('n', None)
        size = kwargs.get('size', n)

        comp = kwargs.get('comp', None)
        shape = kwargs.get('shape', comp)

        max_entry = kwargs.get('max_entry', None)

        # Process positional arguments
        if args:
            # The first arg could be either a size or a shape
            if isinstance(args[0], (int, Integer)):
                if size is not None:
                    raise ValueError("size was specified more than once")
                else:
                    size = args[0]
            else:
                if shape is not None:
                    raise ValueError("the shape was specified more than once")
                shape = args[0]

        # Consistency checks
        if size is not None:
            if not isinstance(size, (int, Integer)):
                raise ValueError("size must be an integer")
            elif size < 0:
                raise ValueError("size must be nonnegative")

        if shape is not None:
            # use in (and not isinstance) below so that lists can be used as
            # shorthand
            if shape not in Compositions():
                raise ValueError("shape must be a composition")
            if any(i == 0 for i in shape):
                raise ValueError("shape must have nonzero parts")
            shape = Composition(shape)

        if (size is not None) and (shape is not None):
            if sum(shape) != size:
                raise ValueError("size and shape are different sizes")

        if max_entry is not None:
            if not isinstance(max_entry, (int, Integer)):
                raise ValueError("max_entry must be an integer")
            elif max_entry <= 0:
                raise ValueError("max_entry must be positive")

        # Dispatch to appropriate class
        if (shape is not None):
            return CompositionTableaux_shape(shape, max_entry)

        if (size is not None):
            return CompositionTableaux_size(size, max_entry)

        return CompositionTableaux_all(max_entry)

    def __init__(self, **kwds):
        r"""
        Initialize ``self``.

        TESTS::

            sage: CT = CompositionTableaux()
            sage: TestSuite(CT).run()
        """
        if 'max_entry' in kwds:
            self.max_entry = kwds['max_entry']
            kwds.pop('max_entry')
        else:
            self.max_entry = None
        super().__init__(**kwds)

    Element = CompositionTableau

    def _element_constructor_(self, t):
        r"""
        Construct an object from ``t`` as an element of ``self``, if
        possible.

        INPUT:

        - ``t`` -- data which can be interpreted as a composition tableau

        OUTPUT: the corresponding CompositionTableau object

        TESTS::

            sage: CT = CompositionTableaux(3)
            sage: CT([[1],[2,2]]).parent() is CT
            True
            sage: CT([[1],[1,2]])
            Traceback (most recent call last):
            ...
            ValueError: [[1], [1, 2]] is not an element of Composition Tableaux of size 3 and maximum entry 3
        """
        if t not in self:
            raise ValueError("%s is not an element of %s" % (t, self))

        return self.element_class(self, t)

    def __contains__(self, T):
        r"""
        Return ``True`` if ``T`` can be interpreted as
        :class:`CompositionTableau`.

        TESTS::

            sage: [[1],[2,2]] in CompositionTableaux(3)
            True
            sage: [[1],[2,2]] in CompositionTableaux(shape=[1,2])
            True
            sage: CompositionTableau([[1],[2,2]]) in CompositionTableaux()
            True
            sage: [[1],[2,2],[2]] in CompositionTableaux()
            False
        """
        if isinstance(T, CompositionTableau):
            return True

        # leftmost column of T strictly increases from top to bottom
        first_col = [row[0] for row in T]
        if any(first_col[i] >= first_col[i+1] for i in range(len(T)-1)):
            return False
        # rows of T weakly decrease from left to right
        for row in T:
            if any(row[i] < row[i+1] for i in range(len(row)-1)):
                return False
        # for 1 <= i < j <= len(comp), for 2 <= k <= m,
        #   T[j,k] \neq 0 and T[j,k] >= T[i,k] ==> T[j,k] > T[i,k-1]
        l = len(T)
        m = max((len(r) for r in T), default=0)
        TT = [row+[0]*(m-len(row)) for row in T]
        for i in range(l):
            for j in range(i+1, l):
                for k in range(1, m):
                    if TT[j][k] != 0 and TT[j][k] >= TT[i][k] and TT[j][k] <= TT[i][k-1]:
                        return False
        return True


class CompositionTableaux_all(CompositionTableaux, DisjointUnionEnumeratedSets):
    r"""
    All composition tableaux.
    """

    def __init__(self, max_entry=None):
        r"""
        Initialize ``self``.

        TESTS::

            sage: CT = CompositionTableaux()
            sage: TestSuite(CT).run()
        """
        self.max_entry = max_entry
        CT_n = lambda n: CompositionTableaux_size(n, max_entry)
        DisjointUnionEnumeratedSets.__init__(self,
            Family(NonNegativeIntegers(), CT_n),
            facade=True, keepkey=False)

    def _repr_(self):
        r"""
        TESTS::

            sage: CompositionTableaux(3)
            Composition Tableaux of size 3 and maximum entry 3

            sage: CompositionTableaux()
            Composition Tableaux
        """
        if self.max_entry is not None:
            return f"Composition tableaux with maximum entry {self.max_entry}"
        return "Composition Tableaux"

    def an_element(self):
        r"""
        Return a particular element of ``self``.

        EXAMPLES::

            sage: CT = CompositionTableaux()
            sage: CT.an_element()
            [[1, 1], [2]]
        """
        return self.element_class(self, [[1, 1], [2]])


class CompositionTableaux_size(CompositionTableaux):
    r"""
    Composition tableaux of a fixed size `n`.

    INPUT:

    - ``n`` -- nonnegative integer
    - ``max_entry`` -- nonnegative integer (default: `n`)

    OUTPUT: the class of composition tableaux of size `n`
    """

    def __init__(self, n, max_entry=None):
        r"""
        Initialize the class of composition tableaux of size `n`.

        TESTS::

            sage: CT = CompositionTableaux(4)
            sage: TestSuite(CT).run()
        """
        if max_entry is None:
            max_entry = n
        super().__init__(max_entry=max_entry,
                         category=FiniteEnumeratedSets())
        self.size = n

    def __contains__(self, x):
        r"""
        TESTS::

            sage: [[1],[2,2]] in CompositionTableaux(3)
            True
            sage: [[1],[2,2]] in CompositionTableaux(4)
            False
        """
        return CompositionTableaux.__contains__(self, x) and sum(map(len, x)) == self.size

    def __iter__(self):
        r"""
        EXAMPLES::

            sage: [t for t in CompositionTableaux(3)]
            [[[1], [2], [3]],
             [[1], [2, 2]],
             [[1], [3, 2]],
             [[1], [3, 3]],
             [[2], [3, 3]],
             [[1, 1], [2]],
             [[1, 1], [3]],
             [[2, 1], [3]],
             [[2, 2], [3]],
             [[1, 1, 1]],
             [[2, 1, 1]],
             [[2, 2, 1]],
             [[2, 2, 2]],
             [[3, 1, 1]],
             [[3, 2, 1]],
             [[3, 2, 2]],
             [[3, 3, 1]],
             [[3, 3, 2]],
             [[3, 3, 3]]]

            sage: CompositionTableaux(3)[0].parent() is CompositionTableaux(3)
            True
        """
        for comp in Compositions(self.size):
            for T in CompositionTableaux_shape(comp, self.max_entry):
                yield self.element_class(self, T)

    def _repr_(self):
        r"""
        TESTS::

            sage: CompositionTableaux(3)
            Composition Tableaux of size 3 and maximum entry 3
        """
        return "Composition Tableaux of size %s and maximum entry %s" % (str(self.size), str(self.max_entry))

    def _an_element_(self):
        r"""
        Return a particular element of ``self``.

        EXAMPLES::

            sage: CT = CompositionTableaux(4)
            sage: CT.an_element()
            [[1, 1, 1], [2]]
            sage: CompositionTableaux(0).an_element()
            []
            sage: CompositionTableaux(1).an_element()
            [[1]]
        """
        if self.size == 0:
            return self.element_class(self, [])
        if self.size == 1:
            return self.element_class(self, [[1]])
        return self.element_class(self, [[1] * (self.size - 1), [2]])


class CompositionTableaux_shape(CompositionTableaux):
    r"""
    Composition tableaux of a fixed shape ``comp`` with a given max entry.

    INPUT:

    - ``comp`` -- a composition
    - ``max_entry`` -- nonnegative integer (default: size of ``comp``)
    """
    def __init__(self, comp, max_entry=None):
        """
        Initialize ``self``.

        TESTS::

            sage: CT = CompositionTableaux([1,2])
            sage: TestSuite(CT).run()

            sage: CT = CompositionTableaux([1,2], max_entry=4)
            sage: TestSuite(CT).run()
        """
        if max_entry is None:
            max_entry = sum(comp)
        super().__init__(max_entry=max_entry,
                         category=FiniteEnumeratedSets())
        self.shape = comp

    def __iter__(self):
        r"""
        An iterator for composition tableaux of a given shape.

        EXAMPLES::

            sage: [t for t in CompositionTableaux([1,2])]
            [[[1], [2, 2]], [[1], [3, 2]], [[1], [3, 3]], [[2], [3, 3]]]
            sage: [t for t in CompositionTableaux([1,2],max_entry=4)]
            [[[1], [2, 2]],
             [[1], [3, 2]],
             [[1], [3, 3]],
             [[1], [4, 2]],
             [[1], [4, 3]],
             [[1], [4, 4]],
             [[2], [3, 3]],
             [[2], [4, 3]],
             [[2], [4, 4]],
             [[3], [4, 4]]]
        """
        if sum(self.shape) == 0:
            yield CompositionTableau([])
        else:
            for z in CompositionTableauxBacktracker(self.shape, self.max_entry):
                yield CompositionTableau(z)

    def __contains__(self, x):
        r"""
        TESTS::

            sage: [[2],[4,3]] in CompositionTableaux([1,2])
            True
            sage: [[2],[3,2]] in CompositionTableaux([1,2])
            False
        """
        return CompositionTableaux.__contains__(self, x) and [len(r) for r in x] == self.shape

    def _repr_(self):
        r"""
        TESTS::

            sage: CompositionTableaux([1,2,1])
            Composition tableaux of shape [1, 2, 1] and maximum entry 4
            sage: CompositionTableaux([1,2,1],max_entry=3)
            Composition tableaux of shape [1, 2, 1] and maximum entry 3
        """
        return "Composition tableaux of shape %s and maximum entry %s" % (str(self.shape), str(self.max_entry))

    def _an_element_(self):
        r"""
        Return a particular element of :class:`CompositionTableaux_shape`.

        EXAMPLES::

            sage: CT = CompositionTableaux([1,2,1])
            sage: CT.an_element()
            [[1], [2, 2], [3]]
        """
        if self.shape == []:
            return self.element_class(self, [])
        t = [[i] * len for i, len in enumerate(self.shape, start=1)]
        return self.element_class(self, t)


class CompositionTableauxBacktracker(GenericBacktracker):
    r"""
    A backtracker class for generating sets of composition tableaux.
    """

    def __init__(self, shape, max_entry=None):
        """
        EXAMPLES::

            sage: from sage.combinat.composition_tableau import CompositionTableauxBacktracker
            sage: n = CompositionTableauxBacktracker([1,3,2])
            sage: n._ending_position
            (2, 1)
            sage: n._initial_state
            (0, 0)
        """
        self._shape = shape
        self._n = sum(shape)
        self._initial_data = [[None] * s for s in shape]
        if max_entry is None:
            max_entry = sum(shape)
        self.max_entry = max_entry

        # The ending position will be at the lowest box which is farthest right
        ending_row = len(shape) - 1
        ending_col = shape[-1] - 1
        self._ending_position = (ending_row, ending_col)

        # Get the highest box that is farthest left
        starting_row = 0
        starting_col = 0

        GenericBacktracker.__init__(self, self._initial_data, (starting_row, starting_col))

    def _rec(self, obj, state):
        r"""
        EXAMPLES::

            sage: from sage.combinat.composition_tableau import CompositionTableauxBacktracker
            sage: n = CompositionTableauxBacktracker([1,3,2])
            sage: obj = [ [None], [None, None, None], [None, None] ]
            sage: list(n._rec(obj, n._initial_state))
            [([[1], [None, None, None], [None, None]], (1, 0), False),
             ([[2], [None, None, None], [None, None]], (1, 0), False),
             ([[3], [None, None, None], [None, None]], (1, 0), False),
             ([[4], [None, None, None], [None, None]], (1, 0), False),
             ([[5], [None, None, None], [None, None]], (1, 0), False),
             ([[6], [None, None, None], [None, None]], (1, 0), False)]
        """
        # Append zeros to a copy of obj
        obj_copy = copy.deepcopy(obj)
        N = max(len(u) for u in obj_copy)
        for a in range(len(obj_copy)):
            Na = len(obj_copy[a])
            obj_copy[a] += [0] * (N - Na)

        # We need to set the i,j^th entry.
        i, j = state

        # Get the next state
        new_state = self.get_next_pos(i, j)
        yld = bool(new_state is None)

        for k in range(1, self.max_entry + 1):
            # We check to make sure that k does not violate the rule weak decrease in rows
            if j != 0 and obj[i][j - 1] < k:
                continue

            # We check to make sure that k does not violate strict increase in first column
            if j == 0 and i != 0 and k <= obj[i - 1][j]:
                continue

            # We check to make sure that k does not violate the Triple Rule
            if j != 0 and i != 0 and any(k == obj_copy[m][j] for m in range(i)):
                continue
            if j != 0 and i != 0 and any(obj_copy[m][j] < k <= obj_copy[m][j - 1]
                                         for m in range(i)):
                continue

            # Fill in the in the i,j box with k
            obj[i][j] = k
            obj_copy[i][j] = k

            # Yield the object
            yield copy.deepcopy(obj), new_state, yld

    def get_next_pos(self, ii, jj):
        r"""
        EXAMPLES::

            sage: from sage.combinat.composition_tableau import CompositionTableauxBacktracker
            sage: T = CompositionTableau([[2,1],[5,4,3,2],[6],[7,7,6]])
            sage: n = CompositionTableauxBacktracker(T.shape_composition())
            sage: n.get_next_pos(1,1)
            (1, 2)
        """
        if (ii, jj) == self._ending_position:
            return None

        for j in range(jj + 1, self._shape[ii]):
            if self._shape[ii] >= j:
                return ii, j

        return ii + 1, 0



from sage.structure.element cimport Element, Matrix, Vector
from sage.structure.parent cimport Parent
from sage.categories.action cimport Action

cdef class MatrixMulAction(Action):
    cdef Parent _codomain
    cdef bint fix_sparseness

cdef class MatrixMatrixAction(MatrixMulAction):
    pass

cdef class MatrixVectorAction(MatrixMulAction):
    pass

cdef class VectorMatrixAction(MatrixMulAction):
    pass

cdef class MatrixPolymapAction(MatrixMulAction):
    pass

cdef class PolymapMatrixAction(MatrixMulAction):
    pass

cdef class MatrixSchemePointAction(MatrixMulAction):
    pass



"""
Cores

A `k`-core is a partition from which no rim hook of size `k` can be removed.
Alternatively, a `k`-core is an integer partition such that the Ferrers
diagram for the partition contains no cells with a hook of size (a
multiple of) `k`.

Authors:

- Anne Schilling and Mike Zabrocki (2011): initial version
- Travis Scrimshaw (2012): Added latex output for Core class
"""
# ****************************************************************************
#       Copyright (C) 2011 Anne Schilling <anne at math.ucdavis.edu>
#                          Mike Zabrocki  <zabrocki at mathstat.yorku.ca>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.parent import Parent
from sage.combinat.partition import Partitions, Partition
from sage.combinat.combinat import CombinatorialElement
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.combinat.combinatorial_map import combinatorial_map


class Core(CombinatorialElement):
    r"""
    A `k`-core is an integer partition from which no rim hook of size `k`
    can be removed.

    EXAMPLES::

        sage: c = Core([2,1],4); c
        [2, 1]
        sage: c = Core([3,1],4); c
        Traceback (most recent call last):
        ...
        ValueError: [3, 1] is not a 4-core
    """
    @staticmethod
    def __classcall_private__(cls, part, k):
        r"""
        Implement the shortcut ``Core(part, k)`` to ``Cores(k,l)(part)``
        where `l` is the length of the core.

        TESTS::

            sage: c = Core([2,1],4); c
            [2, 1]
            sage: c.parent()
            4-Cores of length 3
            sage: type(c)
            <class 'sage.combinat.core.Cores_length_with_category.element_class'>

            sage: Core([2,1],3)
            Traceback (most recent call last):
            ...
            ValueError: [2, 1] is not a 3-core
        """
        if isinstance(part, cls):
            return part
        part = Partition(part)
        if not part.is_core(k):
            raise ValueError("%s is not a %s-core" % (part, k))
        l = sum(part.k_boundary(k).row_lengths())
        return Cores(k, l)(part)

    def __init__(self, parent, core):
        """
        TESTS::

            sage: C = Cores(4,3)
            sage: c = C([2,1]); c
            [2, 1]
            sage: type(c)
            <class 'sage.combinat.core.Cores_length_with_category.element_class'>
            sage: c.parent()
            4-Cores of length 3
            sage: TestSuite(c).run()

            sage: C = Cores(3,3)
            sage: C([2,1])
            Traceback (most recent call last):
            ...
            ValueError: [2, 1] is not a 3-core
        """
        k = parent.k
        part = Partition(core)
        if not part.is_core(k):
            raise ValueError("%s is not a %s-core" % (part, k))
        CombinatorialElement.__init__(self, parent, core)

    def __eq__(self, other) -> bool:
        """
        Test for equality.

        EXAMPLES::

            sage: c = Core([4,2,1,1],5)
            sage: d = Core([4,2,1,1],5)
            sage: e = Core([4,2,1,1],6)
            sage: c == [4,2,1,1]
            False
            sage: c == d
            True
            sage: c == e
            False
        """
        if isinstance(other, Core):
            return (self._list == other._list and
                    self.parent().k == other.parent().k)
        return False

    def __ne__(self, other) -> bool:
        """
        Test for un-equality.

        EXAMPLES::

            sage: c = Core([4,2,1,1],5)
            sage: d = Core([4,2,1,1],5)
            sage: e = Core([4,2,1,1],6)
            sage: c != [4,2,1,1]
            True
            sage: c != d
            False
            sage: c != e
            True
        """
        return not (self == other)

    def __hash__(self):
        """
        Compute the hash of ``self`` by computing the hash of the
        underlying list and of the additional parameter.

        The hash is cached and stored in ``self._hash``.

        EXAMPLES::

            sage: c = Core([4,2,1,1],3)
            sage: c._hash is None
            True
            sage: hash(c) #random
            1335416675971793195
            sage: c._hash #random
            1335416675971793195

        TESTS::

            sage: c = Core([4,2,1,1],5)
            sage: d = Core([4,2,1,1],6)
            sage: hash(c) == hash(d)
            False
        """
        if self._hash is None:
            self._hash = hash(tuple(self._list)) + hash(self.parent().k)
        return self._hash

    def _latex_(self) -> str:
        r"""
        Output the LaTeX representation of this core as a partition.

        See the ``_latex_`` method of :class:`Partition`.

        EXAMPLES::

            sage: c = Core([2,1],4)
            sage: latex(c)
            {\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
            \lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-2}
            \lr{\phantom{x}}\\\cline{1-1}
            \end{array}$}
            }
        """
        return self.to_partition()._latex_()

    def k(self):
        r"""
        Return `k` of the `k`-core ``self``.

        EXAMPLES::

            sage: c = Core([2,1],4)
            sage: c.k()
            4
        """
        return self.parent().k

    @combinatorial_map(name="to partition")
    def to_partition(self):
        r"""
        Turn the core ``self`` into the partition identical to ``self``.

        EXAMPLES::

            sage: mu = Core([2,1,1],3)
            sage: mu.to_partition()
            [2, 1, 1]
        """
        return Partition(self)

    @combinatorial_map(name="to bounded partition")
    def to_bounded_partition(self):
        r"""
        Bijection between `k`-cores and `(k-1)`-bounded partitions.

        This maps the `k`-core ``self`` to the corresponding `(k-1)`-bounded partition.
        This bijection is achieved by deleting all cells in ``self`` of hook length
        greater than `k`.

        EXAMPLES::

            sage: gamma = Core([9,5,3,2,1,1], 5)
            sage: gamma.to_bounded_partition()
            [4, 3, 2, 2, 1, 1]
        """
        k_boundary = self.to_partition().k_boundary(self.k())
        return Partition(k_boundary.row_lengths())

    def size(self):
        r"""
        Return the size of ``self`` as a partition.

        EXAMPLES::

            sage: Core([2,1],4).size()
            3
            sage: Core([4,2],3).size()
            6
        """
        return self.to_partition().size()

    def length(self):
        r"""
        Return the length of ``self``.

        The length of a `k`-core is the size of the corresponding `(k-1)`-bounded partition
        which agrees with the length of the corresponding Grassmannian element,
        see :meth:`to_grassmannian`.

        EXAMPLES::

            sage: c = Core([4,2],3); c.length()
            4
            sage: c.to_grassmannian().length()                                          # needs sage.modules
            4

            sage: Core([9,5,3,2,1,1], 5).length()
            13
        """
        return self.to_bounded_partition().size()

    def to_grassmannian(self):
        r"""
        Bijection between `k`-cores and Grassmannian elements in the affine Weyl group of type `A_{k-1}^{(1)}`.

        For further details, see the documentation of the method
        :meth:`~sage.combinat.partition.Partition.from_kbounded_to_reduced_word` and
        :meth:`~sage.combinat.partition.Partition.from_kbounded_to_grassmannian`.

        EXAMPLES::

            sage: c = Core([3,1,1],3)
            sage: w = c.to_grassmannian(); w                                            # needs sage.modules
            [-1  1  1]
            [-2  2  1]
            [-2  1  2]
            sage: c.parent()
            3-Cores of length 4
            sage: w.parent()                                                            # needs sage.modules
            Weyl Group of type ['A', 2, 1] (as a matrix group acting on the root space)

            sage: c = Core([],3)
            sage: c.to_grassmannian()                                                   # needs sage.modules
            [1 0 0]
            [0 1 0]
            [0 0 1]
        """
        bp = self.to_bounded_partition()
        return bp.from_kbounded_to_grassmannian(self.k() - 1)

    def affine_symmetric_group_simple_action(self, i):
        r"""
        Return the action of the simple transposition `s_i` of the affine symmetric group on ``self``.

        This gives the action of the affine symmetric group of type `A_k^{(1)}` on the `k`-core
        ``self``. If ``self`` has outside (resp. inside) corners of content `i` modulo `k`, then
        these corners are added (resp. removed). Otherwise the action is trivial.

        EXAMPLES::

            sage: c = Core([4,2],3)
            sage: c.affine_symmetric_group_simple_action(0)                             # needs sage.modules
            [3, 1]
            sage: c.affine_symmetric_group_simple_action(1)                             # needs sage.modules
            [5, 3, 1]
            sage: c.affine_symmetric_group_simple_action(2)                             # needs sage.modules
            [4, 2]

        This action corresponds to the left action by the `i`-th simple reflection in the affine
        symmetric group::

            sage: c = Core([4,2],3)
            sage: W = c.to_grassmannian().parent()                                      # needs sage.modules
            sage: i = 0
            sage: (c.affine_symmetric_group_simple_action(i).to_grassmannian()          # needs sage.modules
            ....:     == W.simple_reflection(i)*c.to_grassmannian())
            True
            sage: i = 1
            sage: (c.affine_symmetric_group_simple_action(i).to_grassmannian()          # needs sage.modules
            ....:     == W.simple_reflection(i)*c.to_grassmannian())
            True
        """
        mu = self.to_partition()
        corners = [p for p in mu.outside_corners()
                   if mu.content(p[0], p[1]) % self.k() == i]
        if not corners:
            corners = [p for p in mu.corners()
                       if mu.content(p[0], p[1]) % self.k() == i]
            if not corners:
                return self
            for p in corners:
                mu = mu.remove_cell(p[0])
        else:
            for p in corners:
                mu = mu.add_cell(p[0])
        return Core(mu, self.k())

    def affine_symmetric_group_action(self, w, transposition=False):
        r"""
        Return the (left) action of the affine symmetric group on ``self``.

        INPUT:

        - ``w`` -- tuple of integers `[w_1,\ldots,w_m]` with `0\le w_j<k`.
          If transposition is set to be ``True``, then `w = [w_0,w_1]` is
          interpreted as a transposition `t_{w_0, w_1}`
          (see :meth:`_transposition_to_reduced_word`).

        - ``transposition`` -- boolean (default: ``False``)

        The output is the (left) action of the product of the corresponding simple transpositions
        on ``self``, that is `s_{w_1} \cdots s_{w_m}(self)`. See :meth:`affine_symmetric_group_simple_action`.

        EXAMPLES::

            sage: c = Core([4,2],3)
            sage: c.affine_symmetric_group_action([0,1,0,2,1])
            [8, 6, 4, 2]
            sage: c.affine_symmetric_group_action([0,2], transposition=True)
            [4, 2, 1, 1]

            sage: c = Core([11,8,5,5,3,3,1,1,1],4)
            sage: c.affine_symmetric_group_action([2,5],transposition=True)
            [11, 8, 7, 6, 5, 4, 3, 2, 1]
        """
        c = self
        if transposition:
            w = self._transposition_to_reduced_word(w)
        w.reverse()
        for i in w:
            c = c.affine_symmetric_group_simple_action(i)
        return c

    def _transposition_to_reduced_word(self, t):
        r"""
        Convert the transposition `t = [r,s]` to a reduced word.

        INPUT:

        - ``t`` -- a tuple `[r,s]` such that `r` and `s` are not equivalent mod `k`

        OUTPUT: list of integers in `\{0,1,\ldots,k-1\}` representing a
        reduced word for the transposition `t`

        EXAMPLES::

            sage: c = Core([],4)
            sage: c._transposition_to_reduced_word([2, 5])
            [2, 3, 0, 3, 2]
            sage: c._transposition_to_reduced_word([2, 5]) == c._transposition_to_reduced_word([5,2])
            True
            sage: c._transposition_to_reduced_word([2, 2])
            Traceback (most recent call last):
            ...
            ValueError: t_0 and t_1 cannot be equal mod k

            sage: c = Core([],30)
            sage: c._transposition_to_reduced_word([4, 12])
            [4, 5, 6, 7, 8, 9, 10, 11, 10, 9, 8, 7, 6, 5, 4]

            sage: c = Core([],3)
            sage: c._transposition_to_reduced_word([4, 12])
            [1, 2, 0, 1, 2, 0, 2, 1, 0, 2, 1]
        """
        k = self.k()
        if (t[0] - t[1]) % k == 0:
            raise ValueError("t_0 and t_1 cannot be equal mod k")
        if t[0] > t[1]:
            return self._transposition_to_reduced_word([t[1], t[0]])
        resu = [i % k for i in range(t[0], t[1] - (t[1] - t[0]) // k)]
        resu += [(t[1] - (t[1] - t[0]) // k - 2 - i) % k
                 for i in range(t[1] - (t[1] - t[0]) // k - t[0] - 1)]
        return resu

    def weak_le(self, other):
        r"""
        Weak order comparison on cores.

        INPUT:

        - ``other`` -- another `k`-core

        OUTPUT: boolean

        This returns whether ``self`` <= ``other`` in weak order.

        EXAMPLES::

            sage: c = Core([4,2],3)
            sage: x = Core([5,3,1],3)
            sage: c.weak_le(x)                                                          # needs sage.modules
            True
            sage: c.weak_le([5,3,1])                                                    # needs sage.modules
            True

            sage: x = Core([4,2,2,1,1],3)
            sage: c.weak_le(x)                                                          # needs sage.modules
            False

            sage: x = Core([5,3,1],6)
            sage: c.weak_le(x)
            Traceback (most recent call last):
            ...
            ValueError: the two cores do not have the same k
        """
        if isinstance(other, Core):
            if self.k() != other.k():
                raise ValueError("the two cores do not have the same k")
        else:
            other = Core(other, self.k())
        w = self.to_grassmannian()
        v = other.to_grassmannian()
        return w.weak_le(v, side='left')

    def weak_covers(self):
        r"""
        Return a list of all elements that cover ``self`` in weak order.

        EXAMPLES::

            sage: c = Core([1],3)
            sage: c.weak_covers()                                                       # needs sage.modules
            [[1, 1], [2]]

            sage: c = Core([4,2],3)
            sage: c.weak_covers()                                                       # needs sage.modules
            [[5, 3, 1]]
        """
        w = self.to_grassmannian()
        S = w.upper_covers(side='left')
        S = (x for x in S if x.is_affine_grassmannian())
        return [x.affine_grassmannian_to_core() for x in set(S)]

    def strong_le(self, other):
        r"""
        Strong order (Bruhat) comparison on cores.

        INPUT:

        - ``other`` -- another `k`-core

        OUTPUT: boolean

        This returns whether ``self`` <= ``other`` in Bruhat (or strong) order.

        EXAMPLES::

            sage: c = Core([4,2],3)
            sage: x = Core([4,2,2,1,1],3)
            sage: c.strong_le(x)
            True
            sage: c.strong_le([4,2,2,1,1])
            True

            sage: x = Core([4,1],4)
            sage: c.strong_le(x)
            Traceback (most recent call last):
            ...
            ValueError: the two cores do not have the same k
        """
        if isinstance(other, Core):
            if self.k() != other.k():
                raise ValueError("the two cores do not have the same k")
        else:
            other = Core(other, self.k())
        return other.contains(self)

    def contains(self, other) -> bool:
        r"""
        Check whether ``self`` contains ``other``.

        INPUT:

        - ``other`` -- another `k`-core or a list

        OUTPUT: boolean

        This returns ``True`` if the Ferrers diagram of ``self`` contains the
        Ferrers diagram of ``other``.

        EXAMPLES::

            sage: c = Core([4,2],3)
            sage: x = Core([4,2,2,1,1],3)
            sage: x.contains(c)
            True
            sage: c.contains(x)
            False
        """
        la = self.to_partition()
        mu = Core(other, self.k()).to_partition()
        return la.contains(mu)

    def strong_covers(self):
        r"""
        Return a list of all elements that cover ``self`` in strong order.

        EXAMPLES::

            sage: c = Core([1],3)
            sage: c.strong_covers()
            [[2], [1, 1]]
            sage: c = Core([4,2],3)
            sage: c.strong_covers()
            [[5, 3, 1], [4, 2, 1, 1]]
        """
        S = Cores(self.k(), length=self.length() + 1)
        return [ga for ga in S if ga.contains(self)]

    def strong_down_list(self):
        r"""
        Return a list of all elements that are covered by ``self`` in strong order.

        EXAMPLES::

            sage: c = Core([1],3)
            sage: c.strong_down_list()
            [[]]
            sage: c = Core([5,3,1],3)
            sage: c.strong_down_list()
            [[4, 2], [3, 1, 1]]
        """
        if not self:
            return []
        return [ga for ga in Cores(self.k(), length=self.length() - 1)
                if self.contains(ga)]


def Cores(k, length=None, **kwargs):
    r"""
    A `k`-core is a partition from which no rim hook of size `k` can be removed.
    Alternatively, a `k`-core is an integer partition such that the Ferrers
    diagram for the partition contains no cells with a hook of size (a multiple of) `k`.

    The `k`-cores generally have two notions of size which are
    useful for different applications. One is the number of cells in the
    Ferrers diagram with hook less than `k`, the other is the total
    number of cells of the Ferrers diagram.  In the implementation in
    Sage, the first of notion is referred to as the ``length`` of the `k`-core
    and the second is the ``size`` of the `k`-core.  The class
    of Cores requires that either the size or the length of the elements in
    the class is specified.

    EXAMPLES:

    We create the set of the `4`-cores of length `6`. Here the length of a `k`-core is the size
    of the corresponding `(k-1)`-bounded partition, see also :meth:`~sage.combinat.core.Core.length`::

        sage: C = Cores(4, 6); C
        4-Cores of length 6
        sage: C.list()
        [[6, 3], [5, 2, 1], [4, 1, 1, 1], [4, 2, 2], [3, 3, 1, 1], [3, 2, 1, 1, 1], [2, 2, 2, 1, 1, 1]]
        sage: C.cardinality()
        7
        sage: C.an_element()
        [6, 3]

    We may also list the set of `4`-cores of size `6`, where the size is the number of boxes in the
    core, see also :meth:`~sage.combinat.core.Core.size`::

        sage: C = Cores(4, size=6); C
        4-Cores of size 6
        sage: C.list()
        [[4, 1, 1], [3, 2, 1], [3, 1, 1, 1]]
        sage: C.cardinality()
        3
        sage: C.an_element()
        [4, 1, 1]
    """
    if length is None and 'size' in kwargs:
        return Cores_size(k, kwargs['size'])
    if length is not None:
        return Cores_length(k, length)
    raise ValueError("you need to either specify the length or size of the cores considered")


class Cores_length(UniqueRepresentation, Parent):
    r"""
    The class of `k`-cores of length `n`.
    """

    def __init__(self, k, n):
        """
        TESTS::

            sage: C = Cores(3, 4)
            sage: TestSuite(C).run()
        """
        self.k = k
        self.n = n
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def _repr_(self):
        """
        TESTS::

            sage: repr(Cores(4, 3))  #indirect doctest
            '4-Cores of length 3'
        """
        return "%s-Cores of length %s" % (self.k, self.n)

    def list(self):
        r"""
        Return the list of all `k`-cores of length `n`.

        EXAMPLES::

            sage: C = Cores(3,4)
            sage: C.list()
            [[4, 2], [3, 1, 1], [2, 2, 1, 1]]
        """
        return [la.to_core(self.k - 1)
                for la in Partitions(self.n, max_part=self.k - 1)]

    def from_partition(self, part):
        r"""
        Convert the partition ``part`` into a core (as the identity map).

        This is the inverse method to :meth:`~sage.combinat.core.Core.to_partition`.

        EXAMPLES::

            sage: C = Cores(3,4)
            sage: c = C.from_partition([4,2]); c
            [4, 2]

            sage: mu = Partition([2,1,1])
            sage: C = Cores(3,3)
            sage: C.from_partition(mu).to_partition() == mu
            True

            sage: mu = Partition([])
            sage: C = Cores(3,0)
            sage: C.from_partition(mu).to_partition() == mu
            True
        """
        return Core(part, self.k)

    Element = Core


class Cores_size(UniqueRepresentation, Parent):
    r"""
    The class of `k`-cores of size `n`.
    """

    def __init__(self, k, n):
        """
        TESTS::

            sage: C = Cores(3, size = 4)
            sage: TestSuite(C).run()
        """
        self.k = k
        self.n = n
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def _repr_(self):
        """
        TESTS::

            sage: repr(Cores(4, size = 3))  #indirect doctest
            '4-Cores of size 3'
        """
        return "%s-Cores of size %s" % (self.k, self.n)

    def list(self):
        r"""
        Return the list of all `k`-cores of size `n`.

        EXAMPLES::

            sage: C = Cores(3, size = 4)
            sage: C.list()
            [[3, 1], [2, 1, 1]]
        """
        return [Core(x, self.k) for x in Partitions(self.n)
                if x.is_core(self.k)]

    def from_partition(self, part):
        r"""
        Convert the partition ``part`` into a core (as the identity map).

        This is the inverse method to :meth:`to_partition`.

        EXAMPLES::

            sage: C = Cores(3,size=4)
            sage: c = C.from_partition([2,1,1]); c
            [2, 1, 1]

            sage: mu = Partition([2,1,1])
            sage: C = Cores(3,size=4)
            sage: C.from_partition(mu).to_partition() == mu
            True

            sage: mu = Partition([])
            sage: C = Cores(3,size=0)
            sage: C.from_partition(mu).to_partition() == mu
            True
        """
        return Core(part, self.k)

    Element = Core



# distutils: libraries = flint
# distutils: depends = flint/fmpzi.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fmpzi_init(fmpzi_t x) noexcept
    void fmpzi_clear(fmpzi_t x) noexcept
    void fmpzi_swap(fmpzi_t x, fmpzi_t y) noexcept
    void fmpzi_zero(fmpzi_t x) noexcept
    void fmpzi_one(fmpzi_t x) noexcept
    void fmpzi_set(fmpzi_t res, const fmpzi_t x) noexcept
    void fmpzi_set_si_si(fmpzi_t res, slong a, slong b) noexcept
    void fmpzi_print(const fmpzi_t x) noexcept
    void fmpzi_randtest(fmpzi_t res, flint_rand_t state, mp_bitcnt_t bits) noexcept
    bint fmpzi_equal(const fmpzi_t x, const fmpzi_t y) noexcept
    bint fmpzi_is_zero(const fmpzi_t x) noexcept
    bint fmpzi_is_one(const fmpzi_t x) noexcept
    bint fmpzi_is_unit(const fmpzi_t x) noexcept
    slong fmpzi_canonical_unit_i_pow(const fmpzi_t x) noexcept
    void fmpzi_canonicalise_unit(fmpzi_t res, const fmpzi_t x) noexcept
    slong fmpzi_bits(const fmpzi_t x) noexcept
    void fmpzi_norm(fmpz_t res, const fmpzi_t x) noexcept
    void fmpzi_conj(fmpzi_t res, const fmpzi_t x) noexcept
    void fmpzi_neg(fmpzi_t res, const fmpzi_t x) noexcept
    void fmpzi_add(fmpzi_t res, const fmpzi_t x, const fmpzi_t y) noexcept
    void fmpzi_sub(fmpzi_t res, const fmpzi_t x, const fmpzi_t y) noexcept
    void fmpzi_sqr(fmpzi_t res, const fmpzi_t x) noexcept
    void fmpzi_mul(fmpzi_t res, const fmpzi_t x, const fmpzi_t y) noexcept
    void fmpzi_pow_ui(fmpzi_t res, const fmpzi_t x, ulong exp) noexcept
    void fmpzi_divexact(fmpzi_t q, const fmpzi_t x, const fmpzi_t y) noexcept
    void fmpzi_divrem(fmpzi_t q, fmpzi_t r, const fmpzi_t x, const fmpzi_t y) noexcept
    void fmpzi_divrem_approx(fmpzi_t q, fmpzi_t r, const fmpzi_t x, const fmpzi_t y) noexcept
    slong fmpzi_remove_one_plus_i(fmpzi_t res, const fmpzi_t x) noexcept
    void fmpzi_gcd_euclidean(fmpzi_t res, const fmpzi_t x, const fmpzi_t y) noexcept
    void fmpzi_gcd_euclidean_improved(fmpzi_t res, const fmpzi_t x, const fmpzi_t y) noexcept
    void fmpzi_gcd_binary(fmpzi_t res, const fmpzi_t x, const fmpzi_t y) noexcept
    void fmpzi_gcd_shortest(fmpzi_t res, const fmpzi_t x, const fmpzi_t y) noexcept
    void fmpzi_gcd(fmpzi_t res, const fmpzi_t x, const fmpzi_t y) noexcept
    bint fmpzi_is_prime(const fmpzi_t n) noexcept
    bint fmpzi_is_probabprime(const fmpzi_t n) noexcept



r"""
Morphisms between finitely generated modules over a PID

AUTHOR:

- William Stein, 2009
"""
# *************************************************************************
#       Copyright (C) 2009 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# *************************************************************************

from sage.categories.morphism import Morphism
from .fgp_module import DEBUG
from sage.structure.richcmp import richcmp, op_NE
from sage.misc.cachefunc import cached_method
from sage.categories.homset import Homset
import sage.misc.weak_dict


class FGP_Morphism(Morphism):
    """
    A morphism between finitely generated modules over a PID.

    EXAMPLES:

    An endomorphism::

        sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
        sage: Q = V/W; Q
        Finitely generated module V/W over Integer Ring with invariants (4, 12)
        sage: phi = Q.hom([Q.0+3*Q.1, -Q.1]); phi
        Morphism from module over Integer Ring with invariants (4, 12) to module with invariants (4, 12) that sends the generators to [(1, 3), (0, 11)]
        sage: phi(Q.0) == Q.0 + 3*Q.1
        True
        sage: phi(Q.1) == -Q.1
        True

    A morphism between different modules V1/W1 ---> V2/W2 in
    different ambient spaces::

        sage: V1 = ZZ^2; W1 = V1.span([[1,2],[3,4]]); A1 = V1/W1
        sage: V2 = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W2 = V2.span([2*V2.0+4*V2.1, 9*V2.0+12*V2.1, 4*V2.2]); A2=V2/W2
        sage: A1
        Finitely generated module V/W over Integer Ring with invariants (2)
        sage: A2
        Finitely generated module V/W over Integer Ring with invariants (4, 12)
        sage: phi = A1.hom([2*A2.0])
        sage: phi(A1.0)
        (2, 0)
        sage: 2*A2.0
        (2, 0)
        sage: phi(2*A1.0)
        (0, 0)

    TESTS::

        sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2]); Q = V/W
        sage: phi = Q.hom([Q.0,Q.0 + 2*Q.1])
        sage: loads(dumps(phi)) == phi
        True
    """
    def __init__(self, parent, phi, check=True):
        """
        A morphism between finitely generated modules over a PID.

        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q = V/W; Q
            Finitely generated module V/W over Integer Ring with invariants (4, 12)
            sage: phi = Q.hom([Q.0+3*Q.1, -Q.1]); phi
            Morphism from module over Integer Ring with invariants (4, 12) to module with invariants (4, 12) that sends the generators to [(1, 3), (0, 11)]
            sage: phi(Q.0) == Q.0 + 3*Q.1
            True
            sage: phi(Q.1) == -Q.1
            True

        For full documentation, see :class:`FGP_Morphism`.
        """
        Morphism.__init__(self, parent)
        M = parent.domain()
        N = parent.codomain()
        if isinstance(phi, FGP_Morphism):
            if check:
                if phi.parent() != parent:
                    raise TypeError
            phi = phi._phi
            check = False  # no need

        # input: phi is a morphism from MO = M.optimized().V() to N.V()
        # that sends MO.W() to N.W()
        if check:
            if not isinstance(phi, Morphism) and M == N:
                A = M.optimized()[0].V()
                B = N.V()
                s = M.base_ring()(phi) * B.coordinate_module(A).basis_matrix()
                phi = A.Hom(B)(s)

            MO, _ = M.optimized()
            if phi.domain() != MO.V():
                raise ValueError("domain of phi must be the covering module for the optimized covering module of the domain")
            if phi.codomain() != N.V():
                raise ValueError("codomain of phi must be the covering module the codomain.")
            # check that MO.W() gets sent into N.W()
            # todo (optimize): this is slow:
            for x in MO.W().basis():
                if phi(x) not in N.W():
                    raise ValueError("phi must send optimized submodule of M.W() into N.W()")
        self._phi = phi

    def _repr_(self):
        """
        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q = V/W; Q
            Finitely generated module V/W over Integer Ring with invariants (4, 12)
            sage: phi = Q.hom([Q.0+3*Q.1, -Q.1])
            sage: phi._repr_()
            'Morphism from module over Integer Ring with invariants (4, 12) to module with invariants (4, 12) that sends the generators to [(1, 3), (0, 11)]'
        """
        return "Morphism from module over %s with invariants %s to module with invariants %s that sends the generators to %s" % (
            self.domain().base_ring(), self.domain().invariants(), self.codomain().invariants(),
            list(self.im_gens()))

    @cached_method
    def im_gens(self):
        """
        Return tuple of the images of the generators of the domain
        under this morphism.

        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2]); Q = V/W
            sage: phi = Q.hom([Q.0,Q.0 + 2*Q.1])
            sage: phi.im_gens()
            ((1, 0), (1, 2))
            sage: phi.im_gens() is phi.im_gens()
            True
        """
        return tuple([self(x) for x in self.domain().gens()])

    def _richcmp_(self, right, op):
        """
        Comparison of ``self`` and ``right``.

        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ)
            sage: W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q = V/W
            sage: phi = Q.hom([Q.0,Q.0 + 2*Q.1])
            sage: phi.im_gens()
            ((1, 0), (1, 2))
            sage: phi.im_gens() is phi.im_gens()
            True
            sage: phi == phi
            True
            sage: psi = Q.hom([Q.0,Q.0 - 2*Q.1])
            sage: phi == psi
            False
            sage: psi = Q.hom([Q.0,Q.0 - 2*Q.1])
            sage: phi < psi
            True
            sage: psi >= phi
            True
            sage: psi = Q.hom([Q.0,Q.0 + 2*Q.1])
            sage: phi == psi
            True
        """
        a = (self.domain(), self.codomain())
        b = (right.domain(), right.codomain())
        if a != b:
            return (op == op_NE)
        return richcmp(self.im_gens(), right.im_gens(), op)

    def __add__(self, right):
        """
        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q=V/W; phi = Q.hom([2*Q.0, Q.1]); phi
            Morphism from module over Integer Ring with invariants (4, 12) to module with invariants (4, 12) that sends the generators to [(2, 0), (0, 1)]
            sage: phi + phi
            Morphism from module over Integer Ring with invariants (4, 12) to module with invariants (4, 12) that sends the generators to [(0, 0), (0, 2)]
        """
        if not isinstance(right, FGP_Morphism):  # todo: implement using coercion model
            right = self.parent()(right)
        return FGP_Morphism(self.parent(), self._phi + right._phi, check=DEBUG)

    def __sub__(self, right):
        """
        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q=V/W; phi = Q.hom([2*Q.0, Q.1])
            sage: phi - phi
            Morphism from module over Integer Ring with invariants (4, 12) to module with invariants (4, 12) that sends the generators to [(0, 0), (0, 0)]
        """
        if not isinstance(right, FGP_Morphism):  # todo: implement using coercion model
            right = self.parent()(right)
        return FGP_Morphism(self.parent(), self._phi - right._phi, check=DEBUG)

    def __neg__(self):
        """
        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q=V/W; phi = Q.hom([2*Q.0, Q.1])
            sage: -phi
            Morphism from module over Integer Ring with invariants (4, 12) to module with invariants (4, 12) that sends the generators to [(2, 0), (0, 11)]
        """
        return FGP_Morphism(self.parent(), -self._phi, check=DEBUG)

    def __call__(self, x):
        """
        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q = V/W
            sage: phi = Q.hom([Q.0+3*Q.1, -Q.1])
            sage: phi(Q.0) == Q.0 + 3*Q.1
            True

        We compute the image of some submodules of the domain::

            sage: phi(Q)
            Finitely generated module V/W over Integer Ring with invariants (4, 12)
            sage: phi(Q.submodule([Q.0]))
            Finitely generated module V/W over Integer Ring with invariants (4)
            sage: phi(Q.submodule([Q.1]))
            Finitely generated module V/W over Integer Ring with invariants (12)
            sage: phi(W/W)
            Finitely generated module V/W over Integer Ring with invariants ()

        We try to evaluate on a module that is not a submodule of the domain, which raises a ValueError::

            sage: phi(V/W.scale(2))
            Traceback (most recent call last):
            ...
            ValueError: x must be a submodule or element of the domain

        We evaluate on an element of the domain that is not in the V
        for the optimized representation of the domain::

            sage: V = span([[1/2,0,0],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q = V/W; Q
            Finitely generated module V/W over Integer Ring with invariants (4, 12)
            sage: O, X = Q.optimized()
            sage: O.V()
            Free module of degree 3 and rank 2 over Integer Ring
            User basis matrix:
            [ 0  6  1]
            [ 0 -2  0]
            sage: phi = Q.hom([Q.0, 4*Q.1])
            sage: x = Q(V.0); x
            (0, 8)
            sage: x == 8*Q.1
            True
            sage: x in O.V()
            False
            sage: phi(x)
            (0, 8)
            sage: phi(8*Q.1)
            (0, 8)
            sage: phi(8*Q.1) == phi(x)
            True
        """
        from .fgp_module import FGP_Module_class
        if isinstance(x, FGP_Module_class):
            if not x.is_submodule(self.domain()):
                raise ValueError("x must be a submodule or element of the domain")
            # perhaps can be optimized with a matrix multiply; but note
            # the subtlety of optimized representations.
            return self.codomain().submodule([self(y) for y in x.smith_form_gens()])
        C = self.codomain()
        D = self.domain()
        O, X = D.optimized()
        x = D(x)
        if O is D:
            x = x.lift()
        else:
            # Now we have to transform x so that it is in the optimized representation.
            x = D.V().coordinate_vector(x.lift()) * X
        return C(self._phi(x))

    def kernel(self):
        """
        Compute the kernel of this homomorphism.

        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q = V/W; Q
            Finitely generated module V/W over Integer Ring with invariants (4, 12)
            sage: Q.hom([0, Q.1]).kernel()
            Finitely generated module V/W over Integer Ring with invariants (4)
            sage: A = Q.hom([Q.0, 0]).kernel(); A
            Finitely generated module V/W over Integer Ring with invariants (12)
            sage: Q.1 in A
            True
            sage: phi = Q.hom([Q.0-3*Q.1, Q.0+Q.1])
            sage: A = phi.kernel(); A
            Finitely generated module V/W over Integer Ring with invariants (4)
            sage: phi(A)
            Finitely generated module V/W over Integer Ring with invariants ()
        """
        # The kernel is just got by taking the inverse image of the submodule W
        # of the codomain quotient object.
        V = self._phi.inverse_image(self.codomain().W())
        D = self.domain()
        V = D.W() + V
        return D._module_constructor(V, D.W(), check=DEBUG)

    def inverse_image(self, A):
        """
        Given a submodule A of the codomain of this morphism, return
        the inverse image of A under this morphism.

        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2]); Q = V/W; Q
            Finitely generated module V/W over Integer Ring with invariants (4, 12)
            sage: phi = Q.hom([0, Q.1])
            sage: phi.inverse_image(Q.submodule([]))
            Finitely generated module V/W over Integer Ring with invariants (4)
            sage: phi.kernel()
            Finitely generated module V/W over Integer Ring with invariants (4)
            sage: phi.inverse_image(phi.codomain())
            Finitely generated module V/W over Integer Ring with invariants (4, 12)

            sage: phi.inverse_image(Q.submodule([Q.0]))
            Finitely generated module V/W over Integer Ring with invariants (4)
            sage: phi.inverse_image(Q.submodule([Q.1]))
            Finitely generated module V/W over Integer Ring with invariants (4, 12)

            sage: phi.inverse_image(ZZ^3)
            Traceback (most recent call last):
            ...
            TypeError: A must be a finitely generated quotient module
            sage: phi.inverse_image(ZZ^3 / W.scale(2))
            Traceback (most recent call last):
            ...
            ValueError: A must be a submodule of the codomain
        """
        from .fgp_module import FGP_Module_class
        if not isinstance(A, FGP_Module_class):
            raise TypeError("A must be a finitely generated quotient module")
        if not A.is_submodule(self.codomain()):
            raise ValueError("A must be a submodule of the codomain")
        V = self._phi.inverse_image(A.V())
        D = self.domain()
        V = D.W() + V
        return D._module_constructor(V, D.W(), check=DEBUG)

    def image(self):
        """
        Compute the image of this homomorphism.

        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q = V/W; Q
            Finitely generated module V/W over Integer Ring with invariants (4, 12)
            sage: Q.hom([Q.0+3*Q.1, -Q.1]).image()
            Finitely generated module V/W over Integer Ring with invariants (4, 12)
            sage: Q.hom([3*Q.1, Q.1]).image()
            Finitely generated module V/W over Integer Ring with invariants (12)
        """
        V = self._phi.image() + self.codomain().W()
        W = V.intersection(self.codomain().W())
        return self.codomain()._module_constructor(V, W, check=DEBUG)

    def lift(self, x):
        """
        Given an element x in the codomain of self, if possible find an
        element y in the domain such that self(y) == x.  Raise a ValueError
        if no such y exists.

        INPUT:

        - ``x`` -- element of the codomain of self

        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2])
            sage: Q=V/W; phi = Q.hom([2*Q.0, Q.1])
            sage: phi.lift(Q.1)
            (0, 1)
            sage: phi.lift(Q.0)
            Traceback (most recent call last):
            ...
            ValueError: no lift of element to domain
            sage: phi.lift(2*Q.0)
            (1, 0)
            sage: phi.lift(2*Q.0+Q.1)
            (1, 1)
            sage: V = span([[5, -1/2]],ZZ); W = span([[20,-2]],ZZ); Q = V/W; phi=Q.hom([2*Q.0])
            sage: x = phi.image().0; phi(phi.lift(x)) == x
            True
        """
        x = self.codomain()(x)

        # We view self as a map V/W --> V'/W', where V/W is the
        # optimized representation (which is fine to work with since
        # there is a lift to the optimized representation if and only
        # if there is a lift to the non-optimized representation).
        CD = self.codomain()
        A = self._phi.matrix()
        try:
            H, U = self.__lift_data
        except AttributeError:
            # Get the matrix of self: V --> V' wrt the basis for V and V'.

            # Stack it on top of the basis for W'.
            Wp = CD.V().coordinate_module(CD.W()).basis_matrix()
            B = A.stack(Wp)

            # Compute Hermite form of C with transformation
            H, U = B.hermite_form(transformation=True)
            self.__lift_data = H, U

        # write x in terms of the basis for V.
        w = CD.V().coordinate_vector(x.lift())

        # Solve z*H = w.
        try:
            z = H.solve_left(w)
            if z.denominator() != 1:
                raise ValueError
        except ValueError:
            raise ValueError("no lift of element to domain")

        # Write back in terms of rows of B, and delete rows not corresponding to A,
        # since those corresponding to relations
        v = (z * U)[:A.nrows()]

        # Take the linear combination that v defines.
        y = v * self.domain().optimized()[0].V().basis_matrix()

        # Return the finitely generated module element defined by y.
        y = self.domain()(y)
        assert self(y) == x, "bug in phi.lift()"
        return y


_fgp_homset = sage.misc.weak_dict.WeakValueDictionary()


def FGP_Homset(X, Y):
    """
    EXAMPLES::

        sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2]); Q = V/W
        sage: Q.Hom(Q)           # indirect doctest
        Set of Morphisms from Finitely generated module V/W over Integer Ring with invariants (4, 12) to Finitely generated module V/W over Integer Ring with invariants (4, 12) in Category of modules over Integer Ring
        sage: True # Q.Hom(Q) is Q.Hom(Q)
        True
        sage: type(Q.Hom(Q))
        <class 'sage.modules.fg_pid.fgp_morphism.FGP_Homset_class_with_category'>
    """
    key = (X, Y)
    try:
        return _fgp_homset[key]
    except KeyError:
        pass
    H = FGP_Homset_class(X, Y)
    # Caching breaks tests in fgp_module.
    # _fgp_homset[key] = H
    return H


class FGP_Homset_class(Homset):
    """
    Homsets of :class:`~sage.modules.fg_pid.fgp_module.FGP_Module`.

    TESTS::

        sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2]); Q = V/W
        sage: H = Hom(Q,Q); H    # indirect doctest
        Set of Morphisms from Finitely generated module V/W over Integer Ring with invariants (4, 12) to Finitely generated module V/W over Integer Ring with invariants (4, 12) in Category of modules over Integer Ring
        sage: type(H)
        <class 'sage.modules.fg_pid.fgp_morphism.FGP_Homset_class_with_category'>
    """
    Element = FGP_Morphism

    def __init__(self, X, Y, category=None):
        """
        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2]); Q = V/W
            sage: type(Q.Hom(Q))
            <class 'sage.modules.fg_pid.fgp_morphism.FGP_Homset_class_with_category'>
        """
        if category is None:
            from sage.modules.free_module import FreeModule_generic
            if isinstance(X, FreeModule_generic) and isinstance(Y, FreeModule_generic):
                from sage.categories.modules_with_basis import ModulesWithBasis
                category = ModulesWithBasis(X.base_ring())
            else:
                from sage.categories.modules import Modules
                category = Modules(X.base_ring())
        Homset.__init__(self, X, Y, category)

    def _coerce_map_from_(self, S):
        """
        EXAMPLES::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ); W = V.span([2*V.0+4*V.1, 9*V.0+12*V.1, 4*V.2]); Q = V/W
            sage: phi = Q.hom([Q.0,Q.0 + 2*Q.1]);  psi = loads(dumps(phi))
            sage: phi.parent()._coerce_map_from_(psi.parent())
            True
            sage: phi.parent()._coerce_map_from_(Q.Hom(ZZ^3))
            False
        """
        # We define this so that morphisms in equal parents canonically coerce,
        # since otherwise, e.g., the dumps(loads(...)) doctest above would fail.
        if isinstance(S, FGP_Homset_class) and S == self:
            return True
        if self.is_endomorphism_set():
            R = self.domain().base_ring()
            return R == S or bool(R._coerce_map_from_(S))
        return False

    def __call__(self, x):
        """
        Convert x into an fgp morphism.

        EXAMPLES::

            sage: V = span([[1/2,0,0],[3/2,2,1],[0,0,1]],ZZ); W = V.span([V.0+2*V.1, 9*V.0+2*V.1, 4*V.2])
            sage: Q = V/W; H = Q.Hom(Q)
            sage: H(3)
            Morphism from module over Integer Ring with invariants (4, 16) to module with invariants (4, 16) that sends the generators to [(3, 0), (0, 3)]
        """
        return self.element_class(self, x)



# distutils: libraries = flint
# distutils: depends = flint/qqbar.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void qqbar_init(qqbar_t res) noexcept
    void qqbar_clear(qqbar_t res) noexcept
    qqbar_ptr _qqbar_vec_init(slong len) noexcept
    void _qqbar_vec_clear(qqbar_ptr vec, slong len) noexcept
    void qqbar_swap(qqbar_t x, qqbar_t y) noexcept
    void qqbar_set(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_set_si(qqbar_t res, slong x) noexcept
    void qqbar_set_ui(qqbar_t res, ulong x) noexcept
    void qqbar_set_fmpz(qqbar_t res, const fmpz_t x) noexcept
    void qqbar_set_fmpq(qqbar_t res, const fmpq_t x) noexcept
    void qqbar_set_re_im(qqbar_t res, const qqbar_t x, const qqbar_t y) noexcept
    int qqbar_set_d(qqbar_t res, double x) noexcept
    int qqbar_set_re_im_d(qqbar_t res, double x, double y) noexcept
    slong qqbar_degree(const qqbar_t x) noexcept
    bint qqbar_is_rational(const qqbar_t x) noexcept
    bint qqbar_is_integer(const qqbar_t x) noexcept
    bint qqbar_is_algebraic_integer(const qqbar_t x) noexcept
    bint qqbar_is_zero(const qqbar_t x) noexcept
    bint qqbar_is_one(const qqbar_t x) noexcept
    bint qqbar_is_neg_one(const qqbar_t x) noexcept
    bint qqbar_is_i(const qqbar_t x) noexcept
    bint qqbar_is_neg_i(const qqbar_t x) noexcept
    bint qqbar_is_real(const qqbar_t x) noexcept
    void qqbar_height(fmpz_t res, const qqbar_t x) noexcept
    slong qqbar_height_bits(const qqbar_t x) noexcept
    int qqbar_within_limits(const qqbar_t x, slong deg_limit, slong bits_limit) noexcept
    int qqbar_binop_within_limits(const qqbar_t x, const qqbar_t y, slong deg_limit, slong bits_limit) noexcept
    void _qqbar_get_fmpq(fmpz_t num, fmpz_t den, const qqbar_t x) noexcept
    void qqbar_get_fmpq(fmpq_t res, const qqbar_t x) noexcept
    void qqbar_get_fmpz(fmpz_t res, const qqbar_t x) noexcept
    void qqbar_zero(qqbar_t res) noexcept
    void qqbar_one(qqbar_t res) noexcept
    void qqbar_i(qqbar_t res) noexcept
    void qqbar_phi(qqbar_t res) noexcept
    void qqbar_print(const qqbar_t x) noexcept
    void qqbar_printn(const qqbar_t x, slong n) noexcept
    void qqbar_printnd(const qqbar_t x, slong n) noexcept
    void qqbar_randtest(qqbar_t res, flint_rand_t state, slong deg, slong bits) noexcept
    void qqbar_randtest_real(qqbar_t res, flint_rand_t state, slong deg, slong bits) noexcept
    void qqbar_randtest_nonreal(qqbar_t res, flint_rand_t state, slong deg, slong bits) noexcept
    bint qqbar_equal(const qqbar_t x, const qqbar_t y) noexcept
    bint qqbar_equal_fmpq_poly_val(const qqbar_t x, const fmpq_poly_t f, const qqbar_t y) noexcept
    int qqbar_cmp_re(const qqbar_t x, const qqbar_t y) noexcept
    int qqbar_cmp_im(const qqbar_t x, const qqbar_t y) noexcept
    int qqbar_cmpabs_re(const qqbar_t x, const qqbar_t y) noexcept
    int qqbar_cmpabs_im(const qqbar_t x, const qqbar_t y) noexcept
    int qqbar_cmpabs(const qqbar_t x, const qqbar_t y) noexcept
    int qqbar_cmp_root_order(const qqbar_t x, const qqbar_t y) noexcept
    ulong qqbar_hash(const qqbar_t x) noexcept
    void qqbar_conj(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_re(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_im(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_re_im(qqbar_t res1, qqbar_t res2, const qqbar_t x) noexcept
    void qqbar_abs(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_abs2(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_sgn(qqbar_t res, const qqbar_t x) noexcept
    int qqbar_sgn_re(const qqbar_t x) noexcept
    int qqbar_sgn_im(const qqbar_t x) noexcept
    int qqbar_csgn(const qqbar_t x) noexcept
    void qqbar_floor(fmpz_t res, const qqbar_t x) noexcept
    void qqbar_ceil(fmpz_t res, const qqbar_t x) noexcept
    void qqbar_neg(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_add(qqbar_t res, const qqbar_t x, const qqbar_t y) noexcept
    void qqbar_add_fmpq(qqbar_t res, const qqbar_t x, const fmpq_t y) noexcept
    void qqbar_add_fmpz(qqbar_t res, const qqbar_t x, const fmpz_t y) noexcept
    void qqbar_add_ui(qqbar_t res, const qqbar_t x, ulong y) noexcept
    void qqbar_add_si(qqbar_t res, const qqbar_t x, slong y) noexcept
    void qqbar_sub(qqbar_t res, const qqbar_t x, const qqbar_t y) noexcept
    void qqbar_sub_fmpq(qqbar_t res, const qqbar_t x, const fmpq_t y) noexcept
    void qqbar_sub_fmpz(qqbar_t res, const qqbar_t x, const fmpz_t y) noexcept
    void qqbar_sub_ui(qqbar_t res, const qqbar_t x, ulong y) noexcept
    void qqbar_sub_si(qqbar_t res, const qqbar_t x, slong y) noexcept
    void qqbar_fmpq_sub(qqbar_t res, const fmpq_t x, const qqbar_t y) noexcept
    void qqbar_fmpz_sub(qqbar_t res, const fmpz_t x, const qqbar_t y) noexcept
    void qqbar_ui_sub(qqbar_t res, ulong x, const qqbar_t y) noexcept
    void qqbar_si_sub(qqbar_t res, slong x, const qqbar_t y) noexcept
    void qqbar_mul(qqbar_t res, const qqbar_t x, const qqbar_t y) noexcept
    void qqbar_mul_fmpq(qqbar_t res, const qqbar_t x, const fmpq_t y) noexcept
    void qqbar_mul_fmpz(qqbar_t res, const qqbar_t x, const fmpz_t y) noexcept
    void qqbar_mul_ui(qqbar_t res, const qqbar_t x, ulong y) noexcept
    void qqbar_mul_si(qqbar_t res, const qqbar_t x, slong y) noexcept
    void qqbar_mul_2exp_si(qqbar_t res, const qqbar_t x, slong e) noexcept
    void qqbar_sqr(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_inv(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_div(qqbar_t res, const qqbar_t x, const qqbar_t y) noexcept
    void qqbar_div_fmpq(qqbar_t res, const qqbar_t x, const fmpq_t y) noexcept
    void qqbar_div_fmpz(qqbar_t res, const qqbar_t x, const fmpz_t y) noexcept
    void qqbar_div_ui(qqbar_t res, const qqbar_t x, ulong y) noexcept
    void qqbar_div_si(qqbar_t res, const qqbar_t x, slong y) noexcept
    void qqbar_fmpq_div(qqbar_t res, const fmpq_t x, const qqbar_t y) noexcept
    void qqbar_fmpz_div(qqbar_t res, const fmpz_t x, const qqbar_t y) noexcept
    void qqbar_ui_div(qqbar_t res, ulong x, const qqbar_t y) noexcept
    void qqbar_si_div(qqbar_t res, slong x, const qqbar_t y) noexcept
    void qqbar_scalar_op(qqbar_t res, const qqbar_t x, const fmpz_t a, const fmpz_t b, const fmpz_t c) noexcept
    void qqbar_sqrt(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_sqrt_ui(qqbar_t res, ulong x) noexcept
    void qqbar_rsqrt(qqbar_t res, const qqbar_t x) noexcept
    void qqbar_pow_ui(qqbar_t res, const qqbar_t x, ulong n) noexcept
    void qqbar_pow_si(qqbar_t res, const qqbar_t x, slong n) noexcept
    void qqbar_pow_fmpz(qqbar_t res, const qqbar_t x, const fmpz_t n) noexcept
    void qqbar_pow_fmpq(qqbar_t res, const qqbar_t x, const fmpq_t n) noexcept
    void qqbar_root_ui(qqbar_t res, const qqbar_t x, ulong n) noexcept
    void qqbar_fmpq_root_ui(qqbar_t res, const fmpq_t x, ulong n) noexcept
    void qqbar_fmpq_pow_si_ui(qqbar_t res, const fmpq_t x, slong m, ulong n) noexcept
    int qqbar_pow(qqbar_t res, const qqbar_t x, const qqbar_t y) noexcept
    void qqbar_get_acb(acb_t res, const qqbar_t x, slong prec) noexcept
    void qqbar_get_arb(arb_t res, const qqbar_t x, slong prec) noexcept
    void qqbar_get_arb_re(arb_t res, const qqbar_t x, slong prec) noexcept
    void qqbar_get_arb_im(arb_t res, const qqbar_t x, slong prec) noexcept
    void qqbar_cache_enclosure(qqbar_t res, slong prec) noexcept
    void qqbar_denominator(fmpz_t res, const qqbar_t y) noexcept
    void qqbar_numerator(qqbar_t res, const qqbar_t y) noexcept
    void qqbar_conjugates(qqbar_ptr res, const qqbar_t x) noexcept
    void _qqbar_evaluate_fmpq_poly(qqbar_t res, const fmpz * poly, const fmpz_t den, slong len, const qqbar_t x) noexcept
    void qqbar_evaluate_fmpq_poly(qqbar_t res, const fmpq_poly_t poly, const qqbar_t x) noexcept
    void _qqbar_evaluate_fmpz_poly(qqbar_t res, const fmpz * poly, slong len, const qqbar_t x) noexcept
    void qqbar_evaluate_fmpz_poly(qqbar_t res, const fmpz_poly_t poly, const qqbar_t x) noexcept
    int qqbar_evaluate_fmpz_mpoly_iter(qqbar_t res, const fmpz_mpoly_t poly, qqbar_srcptr x, slong deg_limit, slong bits_limit, const fmpz_mpoly_ctx_t ctx) noexcept
    int qqbar_evaluate_fmpz_mpoly_horner(qqbar_t res, const fmpz_mpoly_t poly, qqbar_srcptr x, slong deg_limit, slong bits_limit, const fmpz_mpoly_ctx_t ctx) noexcept
    int qqbar_evaluate_fmpz_mpoly(qqbar_t res, const fmpz_mpoly_t poly, qqbar_srcptr x, slong deg_limit, slong bits_limit, const fmpz_mpoly_ctx_t ctx) noexcept
    void qqbar_roots_fmpz_poly(qqbar_ptr res, const fmpz_poly_t poly, int flags) noexcept
    void qqbar_roots_fmpq_poly(qqbar_ptr res, const fmpq_poly_t poly, int flags) noexcept
    void qqbar_eigenvalues_fmpz_mat(qqbar_ptr res, const fmpz_mat_t mat, int flags) noexcept
    void qqbar_eigenvalues_fmpq_mat(qqbar_ptr res, const fmpq_mat_t mat, int flags) noexcept
    void qqbar_root_of_unity(qqbar_t res, slong p, ulong q) noexcept
    bint qqbar_is_root_of_unity(slong * p, ulong * q, const qqbar_t x) noexcept
    void qqbar_exp_pi_i(qqbar_t res, slong p, ulong q) noexcept
    void qqbar_cos_pi(qqbar_t res, slong p, ulong q) noexcept
    void qqbar_sin_pi(qqbar_t res, slong p, ulong q) noexcept
    int qqbar_tan_pi(qqbar_t res, slong p, ulong q) noexcept
    int qqbar_cot_pi(qqbar_t res, slong p, ulong q) noexcept
    int qqbar_sec_pi(qqbar_t res, slong p, ulong q) noexcept
    int qqbar_csc_pi(qqbar_t res, slong p, ulong q) noexcept
    int qqbar_log_pi_i(slong * p, ulong * q, const qqbar_t x) noexcept
    int qqbar_atan_pi(slong * p, ulong * q, const qqbar_t x) noexcept
    int qqbar_asin_pi(slong * p, ulong * q, const qqbar_t x) noexcept
    int qqbar_acos_pi(slong * p, ulong * q, const qqbar_t x) noexcept
    int qqbar_acot_pi(slong * p, ulong * q, const qqbar_t x) noexcept
    int qqbar_asec_pi(slong * p, ulong * q, const qqbar_t x) noexcept
    int qqbar_acsc_pi(slong * p, ulong * q, const qqbar_t x) noexcept
    int qqbar_guess(qqbar_t res, const acb_t z, slong max_deg, slong max_bits, int flags, slong prec) noexcept
    int qqbar_express_in_field(fmpq_poly_t res, const qqbar_t alpha, const qqbar_t x, slong max_bits, int flags, slong prec) noexcept
    void qqbar_get_quadratic(fmpz_t a, fmpz_t b, fmpz_t c, fmpz_t q, const qqbar_t x, int factoring) noexcept
    int qqbar_set_fexpr(qqbar_t res, const fexpr_t expr) noexcept
    void qqbar_get_fexpr_repr(fexpr_t res, const qqbar_t x) noexcept
    void qqbar_get_fexpr_root_nearest(fexpr_t res, const qqbar_t x) noexcept
    void qqbar_get_fexpr_root_indexed(fexpr_t res, const qqbar_t x) noexcept
    int qqbar_get_fexpr_formula(fexpr_t res, const qqbar_t x, ulong flags) noexcept
    void qqbar_fmpz_poly_composed_op(fmpz_poly_t res, const fmpz_poly_t A, const fmpz_poly_t B, int op) noexcept
    void qqbar_binary_op(qqbar_t res, const qqbar_t x, const qqbar_t y, int op) noexcept
    int _qqbar_validate_uniqueness(acb_t res, const fmpz_poly_t poly, const acb_t z, slong max_prec) noexcept
    int _qqbar_validate_existence_uniqueness(acb_t res, const fmpz_poly_t poly, const acb_t z, slong max_prec) noexcept
    void _qqbar_enclosure_raw(acb_t res, const fmpz_poly_t poly, const acb_t z, slong prec) noexcept
    void qqbar_enclosure_raw(acb_t res, const qqbar_t x, slong prec) noexcept
    int _qqbar_acb_lindep(fmpz * rel, acb_srcptr vec, slong len, int check, slong prec) noexcept
